from __future__ import annotations

import logging

import aiohttp
from aiohttp import ClientResponse, ClientResponseError, hdrs

logger = logging.getLogger(__name__)


class AuthenticationFailed(ConnectionError):
    pass


class LoginEndpointInvalid(ConnectionError):
    pass


class RealmInvalid(ConnectionError):
    pass


class CountingClientSession(aiohttp.ClientSession):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request_count = 0

    async def _request(self, method, str_or_url, **kwargs):
        self.request_count += 1
        return await super()._request(method, str_or_url, **kwargs)


async def on_request_start(session, context, params: aiohttp.TraceRequestStartParams):
    """
    See https://docs.aiohttp.org/en/stable/tracing_reference.html#aiohttp.TraceConfig.on_request_start
    """
    logger.debug("Starting %s to %s", params.method, params.url)


async def on_request_chunk_sent(
    session, context, params: aiohttp.TraceRequestChunkSentParams
):
    """
    See https://docs.aiohttp.org/en/stable/tracing_reference.html#aiohttp.TraceConfig.on_request_chunk_sent
    """
    if params.chunk:
        logger.debug(
            "Sending %s to %s with %s",
            params.method,
            params.url,
            params.chunk,
        )


async def on_request_end(session, context, params: aiohttp.TraceRequestEndParams):
    """
    See https://docs.aiohttp.org/en/stable/tracing_reference.html#aiohttp.TraceConfig.on_request_end
    and https://docs.python.org/3/howto/logging.html#optimization
    """
    if params.headers.get(hdrs.CONTENT_TYPE, "").lower() == "application/json":
        content = await params.response.json()
    else:
        content = await params.response.text()
    logger.debug(
        "Got response for %s to %s: %s",
        params.method,
        params.url,
        content,
    )


async def on_raise_for_status(response: ClientResponse):
    """
    Add the response text to the exception message of a 400 response
    """
    if response.status == 400:
        text = await response.text()
        try:
            response.raise_for_status()
        except ClientResponseError as e:
            e.message = f"{e.message}, response was: {text}"
            raise e
    response.raise_for_status()


def get_http_client(**kwargs) -> CountingClientSession:
    trace_configs: list[aiohttp.TraceConfig] | None = None
    if logger.isEnabledFor(logging.DEBUG):
        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(on_request_start)
        trace_config.on_request_end.append(on_request_end)
        trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
        trace_configs = [trace_config]

    defaults = dict(
        cookie_jar=aiohttp.CookieJar(),
        raise_for_status=on_raise_for_status,  # type: ignore
        trace_configs=trace_configs,
    )

    return CountingClientSession(**{**defaults, **kwargs})



import pytest as pytest

from ..export import main as export_main
from .utils import list_test_data


@pytest.mark.parametrize("test_data", list_test_data())
async def test_export(mypyllant_aioresponses, test_data) -> None:
    with mypyllant_aioresponses(test_data) as _:
        result = await export_main("test@example.com", "test", "vaillant", "germany")
        assert isinstance(result, list)


@pytest.mark.parametrize("test_data", list_test_data())
async def test_export_data(mypyllant_aioresponses, test_data) -> None:
    with mypyllant_aioresponses(test_data) as _:
        result = await export_main(
            "test@example.com", "test", "vaillant", "germany", data=True
        )
        assert isinstance(result, list)



from __future__ import annotations

import datetime
import json
import logging
import re
from collections.abc import AsyncIterator
from dataclasses import asdict
from html import unescape
from typing import Any
from urllib.parse import parse_qs, urlencode, urlparse
from zoneinfo import ZoneInfo

from aiohttp import ClientResponseError

from myPyllant.const import (
    API_URL_BASE,
    AUTHENTICATE_URL,
    BRANDS,
    CLIENT_ID,
    COUNTRIES,
    DEFAULT_CONTROL_IDENTIFIER,
    DEFAULT_QUICK_VETO_DURATION,
    LOGIN_URL,
    TOKEN_URL,
)
from myPyllant.enums import (
    ControlIdentifier,
    DeviceDataBucketResolution,
    ZoneOperatingModeVRC700,
    ZoneOperatingMode,
    ZoneCurrentSpecialFunction,
    ZoneOperatingType,
    DHWOperationMode,
    VentilationOperationMode,
    VentilationFanStageType,
    DHWOperationModeVRC700,
    DHWCurrentSpecialFunction,
    AmbisenseRoomOperationMode,
    VentilationOperationModeVRC700,
)
from myPyllant.http_client import (
    AuthenticationFailed,
    LoginEndpointInvalid,
    RealmInvalid,
    get_http_client,
    CountingClientSession,
)
from myPyllant.models import (
    Device,
    DeviceData,
    DHWTimeProgram,
    DomesticHotWater,
    Home,
    System,
    SystemReport,
    Ventilation,
    Zone,
    ZoneTimeProgram,
    AmbisenseRoom,
    RoomTimeProgram,
    Circuit,
)
from myPyllant.utils import (
    datetime_format,
    dict_to_camel_case,
    dict_to_snake_case,
    generate_code,
    get_realm,
    get_default_holiday_dates,
)

logger = logging.getLogger(__name__)


def get_system_id(system: System | str) -> str:
    if isinstance(system, System):
        return system.id
    else:
        return system


def get_api_base(control_identifier: ControlIdentifier | str | None = None) -> str:
    key = str(control_identifier) if control_identifier else DEFAULT_CONTROL_IDENTIFIER
    return API_URL_BASE[key]


def get_system_api_base(
    system: str | System, control_identifier: ControlIdentifier | str
) -> str:
    suffix = ""
    if ControlIdentifier(control_identifier) == ControlIdentifier.TLI:
        suffix = "/tli"
    return f"{get_api_base(control_identifier)}/systems/{get_system_id(system)}{suffix}"


class MyPyllantAPI:
    username: str
    password: str
    aiohttp_session: CountingClientSession
    oauth_session: dict = {}
    oauth_session_expires: datetime.datetime | None = None
    control_identifiers: dict[str, str] = {}

    def __init__(
        self, username: str, password: str, brand: str, country: str | None = None
    ) -> None:
        if brand not in BRANDS.keys():
            raise ValueError(
                f"Invalid brand, must be one of {', '.join(BRANDS.keys())}"
            )
        if brand in COUNTRIES:
            # Only need to valid country, if the brand exists as a key in COUNTRIES
            if not country:
                raise RealmInvalid(f"{BRANDS[brand]} requires country to be passed")
            elif country not in COUNTRIES[brand].keys():
                raise RealmInvalid(
                    f"Invalid country, {BRANDS[brand]} only supports {', '.join(COUNTRIES[brand].keys())}"
                )

        self.username = username
        self.password = password
        self.country = country
        self.brand = brand

        self.aiohttp_session = get_http_client()

    async def __aenter__(self) -> MyPyllantAPI:
        try:
            await self.login()
        except Exception:
            await self.aiohttp_session.close()
            raise
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        if not self.aiohttp_session.closed:
            await self.aiohttp_session.close()

    async def login(self):
        code, code_verifier = await self.get_code()
        self.oauth_session = await self.get_token(code, code_verifier)
        logger.debug("Got session %s", self.oauth_session)
        self.set_session_expires()

    async def get_code(self):
        """
        This should really be done in the browser with OIDC, but that's not easy without support from Vaillant

        So instead, we grab the login endpoint from the HTML form of the login website and send username + password
        to obtain a session
        """

        code_verifier, code_challenge = generate_code()
        auth_querystring = {
            "response_type": "code",
            "client_id": CLIENT_ID,
            "code": "code_challenge",
            "redirect_uri": "enduservaillant.page.link://login",
            "code_challenge_method": "S256",
            "code_challenge": code_challenge,
        }

        # Grabbing the login URL from the HTML form of the login page
        code = None
        try:
            async with self.aiohttp_session.get(
                AUTHENTICATE_URL.format(realm=get_realm(self.brand, self.country))
                + "?"
                + urlencode(auth_querystring),
                allow_redirects=False,
            ) as resp:
                login_html = await resp.text()
                if "Location" in resp.headers:
                    parsed_url = urlparse(resp.headers["Location"])
                    code = parse_qs(parsed_url.query).get("code")
        except ClientResponseError as e:
            raise LoginEndpointInvalid from e

        if not code:
            result = re.search(
                LOGIN_URL.format(realm=get_realm(self.brand, self.country))
                + r"\?([^\"]*)",
                login_html,
            )
            login_url = unescape(result.group()) if result else None
            if not login_url:
                raise AuthenticationFailed("Could not get login URL")

            logger.debug("Got login url %s", login_url)
            login_payload = {
                "username": self.username,
                "password": self.password,
                "credentialId": "",
            }
            # Obtaining the code
            async with self.aiohttp_session.post(
                login_url, data=login_payload, allow_redirects=False
            ) as resp:
                logger.debug("Got login response headers %s", resp.headers)
                if "Location" not in resp.headers:
                    raise AuthenticationFailed("Login failed")
                logger.debug(
                    f"Got location from authorize endpoint: {resp.headers['Location']}"
                )
                parsed_url = urlparse(resp.headers["Location"])
                code = parse_qs(parsed_url.query)["code"]
        return code, code_verifier

    async def get_token(self, code, code_verifier):
        # Obtaining access token and refresh token
        token_payload = {
            "grant_type": "authorization_code",
            "client_id": "myvaillant",
            "code": code,
            "code_verifier": code_verifier,
            "redirect_uri": "enduservaillant.page.link://login",
        }

        async with self.aiohttp_session.post(
            TOKEN_URL.format(realm=get_realm(self.brand, self.country)),
            data=token_payload,
            raise_for_status=False,
        ) as resp:
            login_json = await resp.json()
            if resp.status >= 400:
                logger.error(
                    f"Could not log in, got status {resp.status} this response: {login_json}"
                )
                raise Exception(login_json)
            return login_json

    def set_session_expires(self):
        self.oauth_session_expires = datetime.datetime.now(
            datetime.timezone.utc
        ) + datetime.timedelta(seconds=self.oauth_session["expires_in"])
        logger.debug("Session expires in %s", self.oauth_session_expires)

    async def refresh_token(self):
        refresh_payload = {
            "refresh_token": self.oauth_session["refresh_token"],
            "client_id": CLIENT_ID,
            "grant_type": "refresh_token",
        }
        async with self.aiohttp_session.post(
            TOKEN_URL.format(realm=get_realm(self.brand, self.country)),
            data=refresh_payload,
        ) as resp:
            self.oauth_session = await resp.json()
            self.set_session_expires()
            return self.oauth_session

    @property
    def access_token(self):
        return self.oauth_session["access_token"]

    def get_authorized_headers(self):
        return {
            "Authorization": "Bearer " + self.access_token,
            "x-app-identifier": "VAILLANT",
            "Accept-Language": "en-GB",
            "Accept": "application/json, text/plain, */*",
            "x-client-locale": "en-GB",
            "x-idm-identifier": "KEYCLOAK",
            "ocp-apim-subscription-key": "1e0a2f3511fb4c5bbb1c7f9fedd20b1c",
            "User-Agent": "okhttp/4.9.2",
            "Connection": "keep-alive",
        }

    async def get_api_base(
        self,
        system: str | System | None = None,
        control_identifier: ControlIdentifier | str | None = None,
    ) -> str:
        if system and not control_identifier:
            control_identifier = await self.get_control_identifier(system)
        return get_api_base(control_identifier)

    async def get_system_api_base(
        self,
        system: str | System,
        control_identifier: ControlIdentifier | str | None = None,
    ) -> str:
        if not control_identifier:
            control_identifier = await self.get_control_identifier(system)
        return get_system_api_base(system, control_identifier)

    async def get_homes(self) -> AsyncIterator[Home]:
        """
        Returns configured homes and their system IDs

        Returns:
            An Async Iterator with all the configured `Home` objects for the logged-in user
        """
        async with self.aiohttp_session.get(
            f"{await self.get_api_base()}/homes", headers=self.get_authorized_headers()
        ) as homes_resp:
            for home_json in dict_to_snake_case(await homes_resp.json()):
                if "system_id" not in home_json or not home_json["system_id"]:
                    logger.warning(
                        "Skipping home because system_id is missing or empty: %s",
                        home_json,
                    )
                    continue
                timezone = await self.get_time_zone(home_json["system_id"])
                yield Home.from_api(timezone=timezone, **home_json)

    async def get_systems(
        self,
        include_connection_status: bool = False,
        include_diagnostic_trouble_codes: bool = False,
        include_rts: bool = False,
        include_mpc: bool = False,
        include_ambisense_rooms: bool = False,
        include_energy_management: bool = False,
        include_eebus: bool = False,
    ) -> AsyncIterator[System]:
        """
        Returns an async generator of systems under control of the user

        Parameters:
            include_connection_status: Fetches connection status for each system
            include_diagnostic_trouble_codes: Fetches diagnostic trouble codes for each system and device
            include_rts: Fetches RTS data for each system, only supported on TLI controllers
            include_mpc: Fetches MPC data for each system, only supported on TLI controllers
            include_ambisense_rooms: Fetches Ambisense room data
            include_energy_management: Fetches energy management data
            include_eebus: Fetches eebus information

        Returns:
            An Async Iterator with all the `System` objects

        Examples:
            >>> async for system in MyPyllantAPI(**kwargs).get_systems():
            >>>    print(system.water_pressure)
        """
        homes = self.get_homes()
        async for home in homes:
            control_identifier = await self.get_control_identifier(home.system_id)
            system_url = await self.get_system_api_base(home.system_id)
            current_system_url = (
                f"{await self.get_api_base()}/emf/v2/{home.system_id}/currentSystem"
            )

            async with self.aiohttp_session.get(
                system_url, headers=self.get_authorized_headers()
            ) as system_resp:
                system_raw = await system_resp.text()
                if control_identifier.is_vrc700:
                    system_raw = system_raw.replace("domesticHotWater", "dhw")
                    system_raw = system_raw.replace("DomesticHotWater", "Dhw")
                system_json = dict_to_snake_case(json.loads(system_raw))

            async with self.aiohttp_session.get(
                current_system_url, headers=self.get_authorized_headers()
            ) as current_system_resp:
                current_system_json = await current_system_resp.json()

            system = System.from_api(
                brand=self.brand,
                home=home,
                timezone=home.timezone,
                control_identifier=control_identifier,
                connected=await self.get_connection_status(home.system_id)
                if include_connection_status
                else None,
                diagnostic_trouble_codes=await self.get_diagnostic_trouble_codes(
                    home.system_id
                )
                if include_diagnostic_trouble_codes
                else None,
                rts=await self.get_rts(home.system_id) if include_rts else None,
                mpc=await self.get_mpc(home.system_id) if include_mpc else None,
                current_system=dict_to_snake_case(current_system_json),
                ambisense_capability=await self.get_ambisense_capability(
                    home.system_id
                ),
                ambisense_rooms=await self.get_ambisense_rooms(home.system_id)
                if include_ambisense_rooms
                else [],
                energy_management=await self.get_energy_management(home.system_id)
                if include_energy_management
                else None,
                eebus=await self.get_eebus(home.system_id) if include_eebus else None,
                **dict_to_snake_case(system_json),
            )
            yield system

    async def get_data_by_device(
        self,
        device: Device,
        data_resolution: DeviceDataBucketResolution = DeviceDataBucketResolution.DAY,
        data_from: datetime.datetime | None = None,
        data_to: datetime.datetime | None = None,
    ) -> AsyncIterator[DeviceData]:
        """
        Gets all energy data for a device

        Parameters:
            device: The device
            data_resolution: Which resolution level (i.e. day, month)
            data_from: Starting datetime
            data_to: End datetime

        """

        # ISO formatted dates in querystring require timezone information
        if data_from and not data_from.tzinfo:
            data_from = data_from.replace(tzinfo=device.timezone)
        if data_to and not data_to.tzinfo:
            data_to = data_to.replace(tzinfo=device.timezone)

        apis_hit = 0

        for data in device.data:
            if data.skip_data_update:
                logger.debug(
                    "Skipping data update for %s on %s",
                    data.operation_mode,
                    device.name_display,
                )
                yield data
                continue
            data_from = data_from or data.data_from
            if not data_from:
                raise ValueError(
                    "No data_from set, and no data_from found in device data"
                )
            data_to = data_to or data.data_to
            if not data_to:
                raise ValueError("No data_to set, and no data_to found in device data")
            querystring = {
                "resolution": str(data_resolution),
                "operationMode": data.operation_mode,
                "energyType": data.value_type,
                "startDate": data_from.isoformat(timespec="milliseconds"),
                "endDate": data_to.isoformat(timespec="milliseconds"),
            }
            device_buckets_url = (
                f"{await self.get_api_base()}/emf/v2/{device.system_id}/"
                f"devices/{device.device_uuid}/buckets?{urlencode(querystring)}"
            )
            async with self.aiohttp_session.get(
                device_buckets_url, headers=self.get_authorized_headers()
            ) as device_buckets_resp:
                device_buckets_json = await device_buckets_resp.json()
                yield DeviceData.from_api(
                    timezone=device.timezone,
                    device=device,
                    **dict_to_snake_case(device_buckets_json),
                )
                apis_hit += 1
        logger.debug(f"Queried {apis_hit} API endpoints for device data")

    async def get_yearly_reports(
        self,
        system: System,
        year: int | None = None,
    ) -> AsyncIterator[SystemReport]:
        """
        Returns an async generator of systems under control of the user

        Parameters:
            system: The System object or system ID string
            year: The year of the report
        """
        url = f"{await self.get_api_base()}/emf/v2/{system.id}/report/{year}"
        async with self.aiohttp_session.get(
            url, headers=self.get_authorized_headers()
        ) as report_resp:
            reports_json = await report_resp.json()
            for report in dict_to_snake_case(reports_json):
                yield SystemReport.from_api(**report)

    async def set_zone_operating_mode(
        self,
        zone: Zone,
        mode: ZoneOperatingMode | ZoneOperatingModeVRC700 | str,
        operating_type: str = "heating",
    ):
        """
        Sets the operating mode for a zone

        Parameters:
            zone: The target zone
            mode: The target operating mode
            operating_type: Either heating or cooling
        """
        payload = {}
        operating_type = operating_type.lower()
        if operating_type not in ZoneOperatingType:
            raise ValueError(
                f"Invalid HVAC mode, must be one of {', '.join(ZoneOperatingType)}"
            )
        if zone.control_identifier.is_vrc700:
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{operating_type}/operation-mode"
            mode_enum = ZoneOperatingModeVRC700  # type: ignore
        else:
            if operating_type == "cooling":
                url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/operation-mode"
                payload["type"] = operating_type.upper()
            else:
                url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/{operating_type}-operation-mode"
            mode_enum = ZoneOperatingMode  # type: ignore

        if mode not in mode_enum:
            raise ValueError(
                f"Invalid mode, must be one of {', '.join(mode_enum.__members__)}"
            )

        payload["operationMode"] = str(mode)

        await self.aiohttp_session.patch(
            url,
            json=payload,
            headers=self.get_authorized_headers(),
        )

        # zone.heating.operation_mode_heating or zone.cooling.operation_mode_cooling
        setattr(
            getattr(zone, operating_type),
            f"operation_mode_{operating_type}",
            mode_enum(mode),
        )
        return zone

    async def quick_veto_zone_temperature(
        self,
        zone: Zone,
        temperature: float,
        duration_hours: float | None = None,
        default_duration: float | None = None,
        veto_type: str = "heating",
    ):
        """
        Temporarily overwrites the desired temperature in a zone

        Parameters:
            zone: The target zone
            temperature: The target temperature
            duration_hours: Optional, sets overwrite for this many hours
            default_duration: Optional, falls back to this default duration if duration_hours is not given
            veto_type: Only supported on VRC700 controllers, either heating or cooling
        """
        if not default_duration:
            default_duration = DEFAULT_QUICK_VETO_DURATION
        if zone.control_identifier.is_vrc700:
            if veto_type not in ZoneOperatingType:
                raise ValueError(
                    f"Invalid veto type, must be one of {', '.join(ZoneOperatingType)}"
                )
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{veto_type}/quick-veto"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/quick-veto"

        if zone.current_special_function == ZoneCurrentSpecialFunction.QUICK_VETO:
            logger.debug(
                f"Patching quick veto for {zone.name} because it is already in quick veto mode"
            )
            payload = {
                "desiredRoomTemperatureSetpoint": temperature,
            }
            if duration_hours:
                payload["duration"] = duration_hours
            await self.aiohttp_session.patch(
                url,
                json=payload,
                headers=self.get_authorized_headers(),
            )
            zone.desired_room_temperature_setpoint = temperature
            return zone
        else:
            await self.aiohttp_session.post(
                url,
                json={
                    "desiredRoomTemperatureSetpoint": temperature,
                    "duration": duration_hours if duration_hours else default_duration,
                },
                headers=self.get_authorized_headers(),
            )
            zone.desired_room_temperature_setpoint = temperature
            zone.quick_veto_start_date_time = datetime.datetime.now(zone.timezone)
            zone.quick_veto_end_date_time = datetime.datetime.now(
                zone.timezone
            ) + datetime.timedelta(hours=(duration_hours or default_duration))
            return zone

    async def quick_veto_zone_duration(
        self,
        zone: Zone,
        duration_hours: float,
        veto_type: str = "heating",
    ):
        """
        Updates the quick veto duration

        Parameters:
            zone: The target zone
            duration_hours: Updates quick veto duration (in hours)
            veto_type: Only supported on VRC700 controllers, either heating or cooling
        """
        if zone.control_identifier.is_vrc700:
            if veto_type not in ZoneOperatingType:
                raise ValueError(
                    f"Invalid veto type, must be one of {', '.join(ZoneOperatingType)}"
                )
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{veto_type}/quick-veto"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/quick-veto"

        await self.aiohttp_session.patch(
            url,
            json={"duration": duration_hours},
            headers=self.get_authorized_headers(),
        )
        zone.quick_veto_end_date_time = datetime.datetime.now(
            zone.timezone
        ) + datetime.timedelta(hours=duration_hours)
        return zone

    async def set_time_program_temperature(
        self,
        zone: Zone,
        program_type: str,
        temperature: float,
        update_similar_to_dow: str | None = None,
    ):
        logger.debug(f"Setting time program temp {zone.name}")

        if program_type not in ZoneOperatingType:
            raise ValueError(
                "Type must be either heating or cooling, not %s", program_type
            )

        if not zone.heating.time_program_heating:
            raise ValueError(
                "There is no time program set, temperature can't be updated",
                program_type,
            )

        time_program = zone.heating.time_program_heating
        time_program.set_setpoint(temperature, update_similar_to_dow)
        return await self.set_zone_time_program(zone, program_type, time_program)

    async def set_manual_mode_setpoint(
        self,
        zone: Zone,
        temperature: float,
        setpoint_type: str | ZoneOperatingType = "heating",
    ):
        """
        Sets the desired temperature when in manual mode

        Parameters:
            zone: The target zone
            temperature: The target temperature
            setpoint_type: Either heating or cooling
        """
        logger.debug("Setting manual mode setpoint for %s", zone.name)
        if setpoint_type not in ZoneOperatingType:
            raise ValueError(
                f"Invalid veto type, must be one of {', '.join(ZoneOperatingType)}"
            )
        setpoint_type = str(setpoint_type).lower()
        payload: dict[str, Any] = {
            "setpoint": temperature,
        }
        if zone.control_identifier.is_vrc700:
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{setpoint_type}/manual-mode-setpoint"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/manual-mode-setpoint"
            payload["type"] = setpoint_type.upper()
        await self.aiohttp_session.patch(
            url,
            json=payload,
            headers=self.get_authorized_headers(),
        )
        # zone.heating.manual_mode_setpoint_heating or zone.cooling.manual_mode_setpoint_cooling
        setattr(
            getattr(zone, setpoint_type.lower()),
            f"manual_mode_setpoint_{setpoint_type}",
            temperature,
        )

        if (
            setpoint_type == "cooling"
            and zone.cooling
            and zone.cooling.operation_mode_cooling == ZoneOperatingMode.MANUAL
        ):
            zone.desired_room_temperature_setpoint_cooling = temperature

        return zone

    async def set_time_controlled_cooling_setpoint(
        self,
        zone: Zone,
        temperature: float,
    ):
        logger.debug("Setting time controlled setpoint for cooling on %s", zone.name)
        if zone.control_identifier.is_vrc700:
            raise ValueError(
                "Time controlled cooling setpoint is not supported on VRC700 controllers"
            )

        payload: dict[str, Any] = {
            "setpoint": temperature,
        }
        await self.aiohttp_session.patch(
            f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/setpoint-cooling",
            json=payload,
            headers=self.get_authorized_headers(),
        )
        if zone.cooling:
            if zone.cooling.operation_mode_cooling == ZoneOperatingMode.TIME_CONTROLLED:
                zone.desired_room_temperature_setpoint_cooling = temperature
            zone.cooling.setpoint_cooling = temperature
        return zone

    async def cancel_quick_veto_zone_temperature(
        self, zone: Zone, veto_type: str = "heating"
    ):
        """
        Cancels a previously set quick veto in a zone

        Parameters:
            zone: The target zone
            veto_type: Only supported on VRC700 controllers, either heating or cooling
        """
        if zone.control_identifier.is_vrc700:
            if veto_type not in ZoneOperatingType:
                raise ValueError(
                    f"Invalid veto type, must be one of {', '.join(ZoneOperatingType)}"
                )
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{veto_type}/quick-veto"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/quick-veto"

        await self.aiohttp_session.delete(url, headers=self.get_authorized_headers())
        zone.quick_veto_start_date_time = None
        zone.quick_veto_end_date_time = None
        zone.current_special_function = ZoneCurrentSpecialFunction.NONE
        return zone

    async def set_set_back_temperature(
        self, zone: Zone, temperature: float, setback_type: str = "heating"
    ):
        """
        Sets the temperature that a zone gets lowered to in away mode

        Parameters:
            zone: The target zone
            temperature: The setback temperature
            setback_type: Only supported on VRC700 controllers, either heating or cooling
        """
        if zone.control_identifier.is_vrc700:
            if setback_type not in ZoneOperatingType:
                raise ValueError(
                    f"Invalid setback type, must be one of {', '.join(ZoneOperatingType)}"
                )
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{setback_type}/set-back-temperature"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/set-back-temperature"
        await self.aiohttp_session.patch(
            url,
            json={"setBackTemperature": temperature},
            headers=self.get_authorized_headers(),
        )
        # TODO: What to do with cooling?
        if setback_type == "heating":
            zone.heating.set_back_temperature = temperature
        return zone

    async def set_zone_time_program(
        self,
        zone: Zone,
        program_type: str,
        time_program: ZoneTimeProgram,
        setback_type: str = "heating",
    ):
        """
        Sets the temperature that a zone gets lowered to in away mode

        Parameters:
            zone: The target zone
            program_type: Which program to set
            time_program: The time schedule
            setback_type: Only supported on VRC700 controllers, either heating or cooling
        """
        if program_type not in ZoneOperatingType:
            raise ValueError(
                "Type must be either heating or cooling, not %s", program_type
            )
        if zone.control_identifier.is_vrc700:
            if setback_type not in ZoneOperatingType:
                raise ValueError(
                    f"Invalid veto type, must be one of {', '.join(ZoneOperatingType)}"
                )
            url = f"{await self.get_system_api_base(zone.system_id)}/zone/{zone.index}/{setback_type}/time-windows"
        else:
            url = f"{await self.get_system_api_base(zone.system_id)}/zones/{zone.index}/time-windows"
        data = asdict(time_program)
        data["type"] = program_type
        del data["meta_info"]
        await self.aiohttp_session.patch(
            url,
            json=dict_to_camel_case(data),
            headers=self.get_authorized_headers(),
        )

        # zone.heating.time_program_heating = time_program or zone.cooling.time_program_cooling = time_program
        setattr(
            getattr(zone, setback_type), f"time_program_{setback_type}", time_program
        )
        return zone

    async def set_holiday(
        self,
        system: System,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        setpoint: float | None = None,
    ):
        """
        Sets away mode / holiday mode on a system

        Parameters:
            system: The target system
            start: Optional, datetime when the system goes into away mode. Defaults to now
            end: Optional, datetime when away mode should end. Defaults to one year from now
            setpoint: Optional, setpoint temperature during holiday, only supported on VRC700 controllers
        """
        start, end = get_default_holiday_dates(start, end, system.timezone)
        logger.debug(
            "Setting holiday mode for system %s to %s - %s", system.id, start, end
        )
        if not start <= end:
            raise ValueError("Start of holiday mode must be before end")

        data = {
            "startDateTime": datetime_format(start, with_microseconds=True),
            "endDateTime": datetime_format(end, with_microseconds=True),
        }

        if system.control_identifier.is_vrc700:
            url = f"{await self.get_system_api_base(system.id)}/holiday"
            if setpoint is None:
                raise ValueError("setpoint is required on this controller")
            data["setpoint"] = setpoint  # type: ignore
        else:
            url = f"{await self.get_system_api_base(system.id)}/away-mode"
            if setpoint is not None:
                raise ValueError("setpoint is not supported on this controller")

        await self.aiohttp_session.post(
            url, json=data, headers=self.get_authorized_headers()
        )
        for zone in system.zones:
            zone.current_special_function = ZoneCurrentSpecialFunction.HOLIDAY
            zone.general.holiday_start_date_time = start
            zone.general.holiday_end_date_time = end
        return system

    async def cancel_holiday(self, system: System):
        """
        Cancels a previously set away mode / holiday mode on a system

        Parameters:
            system: The target system
        """
        if system.control_identifier.is_vrc700:
            url = f"{await self.get_system_api_base(system.id)}/holiday"
        else:
            url = f"{await self.get_system_api_base(system.id)}/away-mode"

        await self.aiohttp_session.delete(url, headers=self.get_authorized_headers())
        for zone in system.zones:
            zone.current_special_function = ZoneCurrentSpecialFunction.NONE
            zone.general.holiday_start_date_time = None
            zone.general.holiday_end_date_time = None
        return system

    async def set_cooling_for_days(
        self,
        system: System,
        start: datetime.datetime | None = None,
        end: datetime.datetime | None = None,
        duration_days: int | None = None,
    ):
        """
        Sets cooling for a number of days

        Parameters:
            system: The target system
            start: Optional, datetime when the system goes into cooling mode. Defaults to now
            end: Optional, datetime when cooling mode should end. Defaults to one year from now
            duration_days: Optional, number of days to cool
        """
        data: dict[str, int | str] = {}
        if system.control_identifier.is_vrc700:
            if not duration_days:
                raise ValueError("duration_days is required on VRC700 controllers")
            if start or end:
                raise ValueError(
                    "Start and end dates are not supported on VRC700 controllers, use duration_days instead"
                )
            data["value"] = duration_days
        else:
            if duration_days:
                if start or end:
                    raise ValueError(
                        "Start and end dates can't be used together with duration_days"
                    )
                start = datetime.datetime.now(system.timezone)
                end = start + datetime.timedelta(days=duration_days)
            else:
                start, end = get_default_holiday_dates(start, end, system.timezone)
            if not start <= end:
                raise ValueError("Start of holiday mode must be before end")
            data["startDateTime"] = datetime_format(start, with_microseconds=True)
            data["endDateTime"] = datetime_format(end, with_microseconds=True)

        logger.debug(
            "Setting cooling for days on system %s with data %s", system.id, data
        )

        await self.aiohttp_session.post(
            f"{await self.get_system_api_base(system.id)}/cooling-for-days",
            json=data,
            headers=self.get_authorized_headers(),
        )

        if start and end:
            system.configuration["system"]["manual_cooling_start_date"] = (
                datetime_format(start)
            )
            system.configuration["system"]["manual_cooling_end_date"] = datetime_format(
                end
            )
        if system.control_identifier.is_vrc700:
            system.configuration["system"]["cooling_for_x_days"] = duration_days
        return system

    async def cancel_cooling_for_days(self, system: System):
        await self.aiohttp_session.delete(
            f"{await self.get_system_api_base(system.id)}/cooling-for-days",
            headers=self.get_authorized_headers(),
        )
        system.configuration["system"]["manual_cooling_start_date"] = None
        system.configuration["system"]["manual_cooling_end_date"] = None
        return system

    async def set_ventilation_boost(
        self,
        system: System,
    ):
        if system.control_identifier.is_vrc700:
            raise ValueError("Not supported on VRC700 controllers")

        await self.aiohttp_session.post(
            f"{await self.get_system_api_base(system.id)}/ventilation-boost",
            json={},
            headers=self.get_authorized_headers(),
        )

        for zone in system.zones:
            zone.current_special_function = ZoneCurrentSpecialFunction.VENTILATION_BOOST
        return system

    async def cancel_ventilation_boost(self, system: System):
        if system.control_identifier.is_vrc700:
            raise ValueError("Not supported on VRC700 controllers")

        await self.aiohttp_session.delete(
            f"{await self.get_system_api_base(system.id)}/ventilation-boost",
            headers=self.get_authorized_headers(),
        )

        # TODO: Need the switch to the right previous special function
        for zone in system.zones:
            zone.current_special_function = ZoneCurrentSpecialFunction.NONE
        return system

    async def set_domestic_hot_water_temperature(
        self, domestic_hot_water: DomesticHotWater, temperature: int | float
    ):
        """
        Sets the desired hot water temperature

        Parameters:
            domestic_hot_water: The water heater
            temperature: The desired temperature, only whole numbers are supported by the API, floats get rounded
        """
        if isinstance(temperature, float):
            logger.warning("Domestic hot water can only be set to whole numbers")
            temperature = int(round(temperature, 0))
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}"
            f"/domestic-hot-water/{domestic_hot_water.index}/temperature"
        )
        await self.aiohttp_session.patch(
            url, json={"setpoint": temperature}, headers=self.get_authorized_headers()
        )
        domestic_hot_water.tapping_setpoint = temperature
        return domestic_hot_water

    async def boost_domestic_hot_water(self, domestic_hot_water: DomesticHotWater):
        """
        Temporarily boosts hot water temperature

        Parameters:
            domestic_hot_water: The water heater
        """
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}"
            f"/domestic-hot-water/{domestic_hot_water.index}/boost"
        )
        await self.aiohttp_session.post(
            url, json={}, headers=self.get_authorized_headers()
        )
        domestic_hot_water.current_special_function = (
            DHWCurrentSpecialFunction.CYLINDER_BOOST
        )
        return domestic_hot_water

    async def cancel_hot_water_boost(self, domestic_hot_water: DomesticHotWater):
        """
        Cancels hot water boost

        Parameters:
            domestic_hot_water: The water heater
        """
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}"
            f"/domestic-hot-water/{domestic_hot_water.index}/boost"
        )
        await self.aiohttp_session.delete(url, headers=self.get_authorized_headers())
        domestic_hot_water.current_special_function = DHWCurrentSpecialFunction.REGULAR
        return domestic_hot_water

    async def set_domestic_hot_water_operation_mode(
        self,
        domestic_hot_water: DomesticHotWater,
        mode: DHWOperationMode | DHWOperationModeVRC700 | str,
    ):
        """
        Sets the operation mode for water heating

        Parameters:
            domestic_hot_water: The water heater
            mode: The operation mode
        """
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}/domestic-hot-water/"
            f"{domestic_hot_water.index}/operation-mode"
        )
        await self.aiohttp_session.patch(
            url,
            json={"operationMode": str(mode)},
            headers=self.get_authorized_headers(),
        )

        if isinstance(mode, str):
            if domestic_hot_water.control_identifier.is_vrc700:
                mode = DHWOperationModeVRC700(mode)
            else:
                mode = DHWOperationMode(mode)
        domestic_hot_water.operation_mode_dhw = mode
        return domestic_hot_water

    async def set_domestic_hot_water_time_program(
        self, domestic_hot_water: DomesticHotWater, time_program: DHWTimeProgram
    ):
        """
        Sets the schedule for heating water

        Parameters:
            domestic_hot_water: The water heater
            time_program: The schedule
        """
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}"
            f"/domestic-hot-water/{domestic_hot_water.index}/time-windows"
        )
        data = asdict(time_program)
        del data["meta_info"]
        await self.aiohttp_session.patch(
            url,
            json=dict_to_camel_case(data),
            headers=self.get_authorized_headers(),
        )
        domestic_hot_water.time_program_dhw = time_program
        return domestic_hot_water

    async def set_domestic_hot_water_circulation_time_program(
        self, domestic_hot_water: DomesticHotWater, time_program: DHWTimeProgram
    ):
        """
        Sets the schedule for the water circulation pump

        Parameters:
            domestic_hot_water: The water heater
            time_program: The schedule
        """
        url = (
            f"{await self.get_system_api_base(domestic_hot_water.system_id)}"
            f"/domestic-hot-water/{domestic_hot_water.index}/circulation-pump-time-windows"
        )
        data = asdict(time_program)
        del data["meta_info"]
        await self.aiohttp_session.patch(
            url,
            json=dict_to_camel_case(data),
            headers=self.get_authorized_headers(),
        )
        domestic_hot_water.time_program_circulation_pump = time_program
        return domestic_hot_water

    async def set_ventilation_operation_mode(
        self,
        ventilation: Ventilation,
        mode: VentilationOperationMode | VentilationOperationModeVRC700,
    ):
        """
        Sets the operation mode for a ventilation device

        Parameters:
            ventilation: The ventilation device
            mode: The operation mode
        """
        url = (
            f"{await self.get_system_api_base(ventilation.system_id)}"
            f"/ventilation/{ventilation.index}/operation-mode"
        )
        await self.aiohttp_session.patch(
            url,
            json={
                "operationMode": str(mode),
            },
            headers=self.get_authorized_headers(),
        )
        ventilation.operation_mode_ventilation = mode
        return ventilation

    async def set_ventilation_fan_stage(
        self,
        ventilation: Ventilation,
        maximum_fan_stage: int,
        fan_stage_type: VentilationFanStageType,
    ):
        """
        Sets the maximum fan stage for a stage type

        Parameters:
            ventilation: The ventilation device
            maximum_fan_stage: The maximum fan speed, from 1-6
            fan_stage_type: The fan stage type (day or night)
        """
        url = (
            f"{await self.get_system_api_base(ventilation.system_id)}"
            f"/ventilation/{ventilation.index}/fan-stage"
        )
        await self.aiohttp_session.patch(
            url,
            json={
                "maximumFanStage": maximum_fan_stage,
                "type": str(fan_stage_type),
            },
            headers=self.get_authorized_headers(),
        )
        setattr(
            ventilation,
            f"maximum_{fan_stage_type.lower()}_fan_stage",
            maximum_fan_stage,
        )
        return ventilation

    async def get_connection_status(self, system: System | str) -> bool:
        """
        Returns whether the system is online

        Parameters:
            system: The System object or system ID string
        """
        url = (
            f"{await self.get_api_base()}/systems/"
            f"{get_system_id(system)}/meta-info/connection-status"
        )
        response = await self.aiohttp_session.get(
            url,
            headers=self.get_authorized_headers(),
        )
        try:
            return (await response.json())["connected"]
        except KeyError:
            logger.warning("Couldn't get connection status")
            return False

    async def get_control_identifier(self, system: System | str) -> ControlIdentifier:
        """
        The control identifier is used in the URL to request system information (usually `tli`)

        Parameters:
            system: The System object or system ID string
        """
        system_id = get_system_id(system)

        if system_id in self.control_identifiers:
            # We already have the control identifier cached
            control_identifier = self.control_identifiers[system_id]
        else:
            url = (
                f"{await self.get_api_base()}/systems/"
                f"{system_id}/meta-info/control-identifier"
            )
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
            try:
                control_identifier = (await response.json())["controlIdentifier"]
                self.control_identifiers[system_id] = control_identifier
            except KeyError:
                logger.warning("Couldn't get control identifier")
                control_identifier = DEFAULT_CONTROL_IDENTIFIER

        return ControlIdentifier(control_identifier)

    async def get_time_zone(self, system: System | str) -> datetime.tzinfo | None:
        """
        Gets the configured timezone for a system

        Parameters:
            system: The System object or system ID string
        """
        url = (
            f"{await self.get_api_base()}/systems/"
            f"{get_system_id(system)}/meta-info/time-zone"
        )
        response = await self.aiohttp_session.get(
            url,
            headers=self.get_authorized_headers(),
        )
        try:
            tz = (await response.json())["timeZone"]
            return ZoneInfo(key=tz)
        except (KeyError, TypeError):
            logger.warning("Couldn't get timezone from API")
            return None

    async def get_diagnostic_trouble_codes(
        self, system: System | str
    ) -> list[dict] | None:
        """
        Returns a list of trouble codes by device

        Parameters:
            system: The System object or system ID string
        """
        url = (
            f"{await self.get_api_base()}/systems/"
            f"{get_system_id(system)}/diagnostic-trouble-codes"
        )
        try:
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get diagnostic trouble codes", exc_info=e)
            return None
        result = await response.json()
        return dict_to_snake_case(result)

    async def get_rts(self, system: System | str) -> dict:
        """
        Gets RTS data, which contains on/off cycles and operation time

        Parameters:
            system: The System object or system ID string
        """
        url = f"{await self.get_api_base()}/rts/{get_system_id(system)}/devices"
        try:
            logger.debug("Getting RTS data")
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get RTS data", exc_info=e)
            return {"statistics": []}
        result = await response.json()
        return dict_to_snake_case(result)

    async def get_mpc(self, system: System | str) -> dict:
        """
        Gets live power usage data per device

        Parameters:
            system: The System object or system ID string
        """
        url = f"{await self.get_api_base()}/hem/{get_system_id(system)}/mpc"
        try:
            logger.debug("Getting MPC data")
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get MPC data", exc_info=e)
            return {"devices": []}
        result = await response.json()
        return dict_to_snake_case(result)

    async def get_energy_management(self, system: System | str) -> dict:
        """
        Gets energy management information

        Parameters:
            system: The System object or system ID string
        """
        url = f"{await self.get_api_base()}/eebus/energy-management/{get_system_id(system)}"
        try:
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get energy management data", exc_info=e)
            return {}
        result = await response.json()
        return dict_to_snake_case(result)

    async def get_eebus(self, system: System | str) -> dict:
        """
        Gets EEBUS information

        Parameters:
            system: The System object or system ID string
        """
        url = f"{await self.get_api_base()}/ship/{get_system_id(system)}/self"
        try:
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get eebus information", exc_info=e)
            return {}
        result = await response.json()
        return dict_to_snake_case(result)

    async def toggle_eebus(
        self, system: System | str, enabled: bool = True
    ) -> System | None:
        """
        Enables EEBUS interface

        Parameters:
            system: The System object or system ID string
            enabled: Whether to enable or disable EEBUS
        """
        await self.aiohttp_session.put(
            f"{await self.get_api_base()}/ship/{get_system_id(system)}/self/spine",
            json={"enabled": enabled},
            headers=self.get_authorized_headers(),
        )
        if isinstance(system, System) and system.eebus:
            system.eebus["spline_enabled"] = enabled
            return system
        return None

    async def get_ambisense_capability(self, system: System | str) -> bool:
        """
        Whether a system is ambisense capable

        Parameters:
            system: The System object or system ID string
        """
        url = f"{get_api_base()}/api/v1/ambisense/facilities/{get_system_id(system)}/capability"
        try:
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get ambisense capability data", exc_info=e)
            return False
        result = await response.json()
        return dict_to_snake_case(result).get("rbr_capable", False)

    async def get_ambisense_rooms(self, system: System | str) -> list[dict]:
        """
        Whether a system is ambisense capable

        Parameters:
            system: The System object or system ID string
        """
        url = f"{get_api_base()}/api/v1/ambisense/facilities/{get_system_id(system)}/rooms"
        try:
            response = await self.aiohttp_session.get(
                url,
                headers=self.get_authorized_headers(),
            )
        except ClientResponseError as e:
            logger.warning("Could not get rooms data", exc_info=e)
            return []
        result = dict_to_snake_case(await response.json())
        for room in result:
            room["time_program"] = room.pop("timeprogram")
        return result

    async def set_ambisense_room_operation_mode(
        self,
        room: AmbisenseRoom,
        mode: AmbisenseRoomOperationMode | str,
    ) -> AmbisenseRoom:
        """
        Sets the operation mode for a room

        Parameters:
            room: The room
            mode: The operation mode
        """
        url = f"{await self.get_api_base()}/api/v1/ambisense/facilities/{room.system_id}/rooms/{room.room_index}/configuration/operation-mode"
        await self.aiohttp_session.put(
            url,
            json={"operationMode": str(mode).lower()},
            headers=self.get_authorized_headers(),
        )

        if isinstance(mode, str):
            room.room_configuration.operation_mode = AmbisenseRoomOperationMode(
                mode.upper()
            )
        else:
            room.room_configuration.operation_mode = mode
        return room

    async def quick_veto_ambisense_room(
        self,
        room: AmbisenseRoom,
        temperature: float,
        duration_minutes: int | None = None,
        default_duration: int | None = None,
    ) -> AmbisenseRoom:
        """
        Temporarily overwrites the desired temperature in a room

        Parameters:
            room: The target room
            temperature: The target temperature
            duration_minutes: Optional, sets overwrite for this many minutes
            default_duration: Optional, falls back to this default duration if duration_minutes is not given
        """
        if not default_duration:
            default_duration = (
                int(DEFAULT_QUICK_VETO_DURATION) * 60
            )  # duration for quick veto for room is in minutes

        if duration_minutes and duration_minutes < 30:
            raise ValueError("duration_minutes must be greater than 30")

        url = f"{await self.get_api_base()}/api/v1/ambisense/facilities/{room.system_id}/rooms/{room.room_index}/configuration/quick-veto"

        payload = {
            "temperatureSetpoint": temperature,
            "duration": duration_minutes or default_duration,
        }

        await self.aiohttp_session.put(
            url,
            json=payload,
            headers=self.get_authorized_headers(),
        )

        room.room_configuration.temperature_setpoint = temperature
        room.room_configuration.quick_veto_end_time = datetime.datetime.now(
            datetime.timezone.utc
        ) + datetime.timedelta(minutes=(duration_minutes or default_duration))
        return room

    async def cancel_quick_veto_ambisense_room(
        self, room: AmbisenseRoom
    ) -> AmbisenseRoom:
        """
        Cancels a previously set quick veto in a room

        Parameters:
            room: The target room
        """
        url = f"{await self.get_api_base()}/api/v1/ambisense/facilities/{room.system_id}/rooms/{room.room_index}/configuration/quick-veto"

        await self.aiohttp_session.delete(url, headers=self.get_authorized_headers())
        room.room_configuration.quick_veto_end_time = None
        return room

    async def set_ambisense_room_manual_mode_setpoint_temperature(
        self,
        room: AmbisenseRoom,
        temperature: float,
    ):
        """
        Sets the desired temperature when in manual mode. The temperature is only taken into account if the room is in
        MANUAL mode, otherwise it has no effect.

        Parameters:
            room: The target room
            temperature: The target temperature
        """
        logger.debug(
            "Setting manual mode setpoint temperature to %.1f for %s",
            temperature,
            room.name,
        )
        payload: dict[str, Any] = {
            "temperatureSetpoint": temperature,
        }
        url = f"{await self.get_api_base()}/api/v1/ambisense/facilities/{room.system_id}/rooms/{room.room_index}/configuration/temperature-setpoint"

        await self.aiohttp_session.put(
            url,
            json=payload,
            headers=self.get_authorized_headers(),
        )
        room.room_configuration.temperature_setpoint = temperature
        return room

    async def set_ambisense_room_time_program(
        self, room: AmbisenseRoom, time_program: RoomTimeProgram
    ) -> AmbisenseRoom:
        """
        Set time program for an ambisense room

        Parameters:
            room: The target room
            time_program: The new time program
        """
        url = f"{await self.get_api_base()}/api/v1/ambisense/facilities/{room.system_id}/rooms/{room.room_index}/timeprogram"

        data = asdict(time_program, dict_factory=RoomTimeProgram.dict_factory)
        payload = dict_to_camel_case(data)

        await self.aiohttp_session.put(
            url, json=payload, headers=self.get_authorized_headers()
        )
        room.time_program = time_program
        return room

    async def set_circuit_heating_curve(
        self,
        circuit: Circuit,
        heating_curve: float,
    ) -> Circuit:
        """
        Set heating curve value for a circuit

        :param circuit:
        :param heating_curve:
        :return:
        """
        url = f"{await self.get_system_api_base(circuit.system_id)}/circuit/{circuit.index}/heating-curve"

        await self.aiohttp_session.patch(
            url,
            json={"heatingCurve": heating_curve},
            headers=self.get_authorized_headers(),
        )
        circuit.heating_curve = heating_curve
        return circuit

    async def set_circuit_heat_demand_limited_by_outside_temperature(
        self,
        circuit: Circuit,
        heat_demand_limited_by_outside_temperature: float,
    ) -> Circuit:
        """
        Set heating curve value for a circuit

        :param circuit:
        :param heat_demand_limited_by_outside_temperature:
        :return:
        """
        url = f"{await self.get_system_api_base(circuit.system_id)}/circuit/{circuit.index}/heat-demand-limited-by-outside-temperature"

        await self.aiohttp_session.post(
            url,
            json={
                "heatDemandLimitedByOutsideTemperature": heat_demand_limited_by_outside_temperature
            },
            headers=self.get_authorized_headers(),
        )
        circuit.heat_demand_limited_by_outside_temperature = (
            heat_demand_limited_by_outside_temperature
        )
        return circuit

    async def set_circuit_min_flow_temperature_setpoint(
        self,
        circuit: Circuit,
        min_flow_temperature_setpoint: float,
    ) -> Circuit:
        """
        Set heating curve value for a circuit

        :param circuit:
        :param min_flow_temperature_setpoint:
        :return:
        """
        url = f"{await self.get_system_api_base(circuit.system_id)}/circuit/{circuit.index}/min-flow-temperature-setpoint"

        await self.aiohttp_session.patch(
            url,
            json={"minFlowTemperatureSetpoint": min_flow_temperature_setpoint},
            headers=self.get_authorized_headers(),
        )
        circuit.heating_flow_temperature_minimum_setpoint = (
            min_flow_temperature_setpoint
        )
        return circuit



#!/usr/bin/env python3

import argparse
import asyncio
import logging
from dataclasses import asdict
from datetime import datetime, timedelta

from myPyllant.api import MyPyllantAPI
from myPyllant.models import MyPyllantDataClass
from myPyllant.utils import add_default_parser_args, recursive_compare

parser = argparse.ArgumentParser(
    description="Test service calls with real API, without changing anything."
)
add_default_parser_args(parser)
parser.add_argument(
    "-v", "--verbose", help="increase output verbosity", action="store_true"
)


async def print_changes(func, object: MyPyllantDataClass, *args, **kwargs):
    old = asdict(object)
    new: MyPyllantDataClass = await func(object, *args, **kwargs)
    print(f"{func.__name__} changes:")
    recursive_compare(old, asdict(new))
    print()


async def main(user, password, brand, country):
    async with MyPyllantAPI(user, password, brand, country) as api:
        async for system in api.get_systems(
            include_connection_status=True,
            include_diagnostic_trouble_codes=True,
        ):
            if system.control_identifier.is_vrc700:
                await print_changes(
                    api.set_holiday,
                    system,
                    datetime.now(system.timezone),
                    datetime.now(system.timezone) + timedelta(days=7),
                    10.0,
                )
            else:
                await print_changes(
                    api.set_holiday,
                    system,
                    datetime.now(system.timezone),
                    datetime.now(system.timezone) + timedelta(days=7),
                )
            await print_changes(api.cancel_holiday, system)
            if system.zones:
                zone = system.zones[0]
                await print_changes(
                    api.set_set_back_temperature,
                    zone,
                    zone.heating.set_back_temperature,
                )
                if zone.heating.manual_mode_setpoint_heating:
                    await print_changes(
                        api.set_manual_mode_setpoint,
                        zone,
                        zone.heating.manual_mode_setpoint_heating,
                    )
                await print_changes(
                    api.set_zone_operating_mode,
                    zone,
                    zone.heating.operation_mode_heating,
                )
                await print_changes(api.quick_veto_zone_temperature, zone, 21, 5)
                await print_changes(api.cancel_quick_veto_zone_temperature, zone)
                if zone.heating.time_program_heating:
                    await print_changes(
                        api.set_zone_time_program,
                        zone,
                        "heating",
                        zone.heating.time_program_heating,
                    )

                    if zone.heating.time_program_heating.monday[0].setpoint:
                        await print_changes(
                            api.set_time_program_temperature,
                            zone,
                            "heating",
                            zone.heating.time_program_heating.monday[0].setpoint,
                        )

            if system.domestic_hot_water:
                dhw = system.domestic_hot_water[0]
                await print_changes(
                    api.set_domestic_hot_water_operation_mode,
                    dhw,
                    dhw.operation_mode_dhw,
                )
                await print_changes(api.boost_domestic_hot_water, dhw)
                await print_changes(api.cancel_hot_water_boost, dhw)

                if dhw.tapping_setpoint:
                    await print_changes(
                        api.set_domestic_hot_water_temperature,
                        dhw,
                        int(dhw.tapping_setpoint),
                    )
                await print_changes(
                    api.set_domestic_hot_water_time_program, dhw, dhw.time_program_dhw
                )
                await print_changes(
                    api.set_domestic_hot_water_circulation_time_program,
                    dhw,
                    dhw.time_program_circulation_pump,
                )

            if system.circuits:
                circuit = system.circuits[0]
                await print_changes(
                    api.set_circuit_heating_curve,
                    circuit,
                    circuit.heating_curve,
                )
                await print_changes(
                    api.set_circuit_heat_demand_limited_by_outside_temperature,
                    circuit,
                    circuit.heat_demand_limited_by_outside_temperature,
                )
                await print_changes(
                    api.set_circuit_min_flow_temperature_setpoint,
                    circuit,
                    circuit.heating_flow_temperature_minimum_setpoint,
                )


if __name__ == "__main__":
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    asyncio.run(main(args.user, args.password, args.brand, args.country))



import json
import re
from collections import defaultdict
from pathlib import Path
from typing import Any

import yaml
from aioresponses import CallbackResult, aioresponses

from myPyllant.api import MyPyllantAPI
from myPyllant.const import API_URL_BASE, LOGIN_URL
from myPyllant.tests.generate_test_data import DATA_DIR
from myPyllant.utils import get_realm


def _mypyllant_aioresponses():
    class _mypyllant_aioresponses(aioresponses):
        def __init__(
            self, test_data=None, raise_exception: Exception | None = None, **kwargs
        ):
            self.test_data = test_data
            self.test_exception = raise_exception
            super().__init__(**kwargs)

        def __enter__(self):
            super().__enter__()

            if self.test_exception:
                self.get(
                    re.compile(r".*"),
                    exception=self.test_exception,
                )
                self.post(
                    re.compile(r".*"),
                    exception=self.test_exception,
                )
                self.patch(
                    re.compile(r".*"),
                    exception=self.test_exception,
                )
                self.delete(
                    re.compile(r".*"),
                    exception=self.test_exception,
                )

            # auth endpoints
            self.get(
                re.compile(r".*vaillant-germany-b2c/protocol/openid-connect/auth\?"),
                body=f"{LOGIN_URL.format(realm=get_realm('vaillant', 'germany'))}?test=test",
                status=200,
                repeat=True,
            )
            self.get(
                re.compile(r".*bulex-b2c/protocol/openid-connect/auth\?"),
                body=f"{LOGIN_URL.format(realm=get_realm('bulex'))}?test=test",
                status=200,
                repeat=True,
            )
            self.get(
                re.compile(r".*glow-worm-b2c/protocol/openid-connect/auth\?"),
                body=f"{LOGIN_URL.format(realm=get_realm('glow-worm'))}?test=test",
                status=200,
                repeat=True,
            )
            self.post(
                re.compile(r".*login-actions/authenticate\?"),
                status=200,
                headers={"Location": "test?code=code"},
                repeat=True,
            )
            self.post(
                re.compile(r".*openid-connect/token$"),
                status=200,
                payload={
                    "expires_in": 3600,
                    "access_token": "access_token",
                    "refresh_token": "refresh_token",
                },
                repeat=True,
            )
            # API endpoints
            actions = re.compile(
                r".*(away-mode|setBackTemperature|temperature|holiday|cooling-for-days|ventilation-boost)$"
            )
            self.post(
                actions,
                status=200,
                payload={},
                repeat=True,
            )
            self.patch(
                actions,
                status=200,
                payload={},
                repeat=True,
            )
            self.delete(
                actions,
                status=200,
                payload={},
                repeat=True,
            )
            self.post(
                re.compile(r".*domestic-hot-water/.*/temperature$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.patch(
                re.compile(r".*domestic-hot-water/.*/operation-mode$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.post(
                re.compile(r".*domestic-hot-water/.*/boost$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.delete(
                re.compile(r".*domestic-hot-water/.*/boost$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.post(
                re.compile(
                    r".*zones?/.*/(quick-veto|manual-mode-setpoint|heating-operation-mode|operation-mode|setpoint-cooling)$"
                ),
                status=200,
                payload={},
                repeat=True,
            )
            self.patch(
                re.compile(
                    r".*zones?/.*/(quick-veto|manual-mode-setpoint|heating-operation-mode|operation-mode|setpoint-cooling)$"
                ),
                status=200,
                payload={},
                repeat=True,
            )
            self.delete(
                re.compile(r".*zones?/.*/quick-veto$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.patch(
                re.compile(r".*ventilation/.*/(operation-mode|fan-stage)$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.put(
                re.compile(
                    r".*/ambisense/.*/rooms/.*/configuration/(operation-mode|quick-veto|temperature-setpoint)$"
                ),
                status=200,
                payload={},
                repeat=True,
            )
            self.put(
                re.compile(r".*/ambisense/.*/rooms/.*/timeprogram$"),
                status=200,
                payload={},
                repeat=True,
            )
            self.delete(
                re.compile(r".*/ambisense/.*/rooms/.*/configuration/quick-veto$"),
                status=200,
                payload={},
                repeat=True,
            )

            def get_test_data(url: str, key: str, default=None) -> CallbackResult:
                """
                Return test data CallbackResult based on the URL and key
                """
                url_parts = None
                for api_base in API_URL_BASE.values():
                    if url.startswith(api_base):
                        url_parts = url.replace(api_base, "").split("/")
                        break
                if not url_parts:
                    raise ValueError(f"Could not find API base in URL {url}")
                if url_parts[1] == "emf":
                    system_id = url_parts[3]
                elif "ambisense/facilities" in url:
                    system_id = url_parts[5]
                else:
                    system_id = url_parts[2]
                return CallbackResult(
                    status=200, payload=self.test_data[system_id].get(key, default)
                )

            def test_data_by_system(url, **kwargs):
                """
                Return test data based on the system ID in the URL
                """
                url = str(url)
                result = None

                match url:
                    case url if re.match(r".*currentSystem$", url):
                        result = get_test_data(url, "current_system")
                    case url if re.match(r".*buckets\?.*", url):
                        result = get_test_data(url, "device_buckets")
                    case url if re.match(r".*systems/.*/tli", url):
                        result = get_test_data(url, "system")
                    case url if re.match(r".*vrc700.*systems.*", url):
                        result = get_test_data(url, "system")
                    case url if re.match(r".*meta-info/control-identifier$", url):
                        result = get_test_data(url, "control_identifier")
                    case url if re.match(r".*meta-info/time-zone", url):
                        result = get_test_data(url, "time_zone")
                    case url if re.match(r".*meta-info/connection-status", url):
                        result = get_test_data(url, "connection_status")
                    case url if re.match(r".*diagnostic-trouble-codes$", url):
                        result = get_test_data(url, "diagnostic_trouble_codes", [])
                    case url if re.match(r".*/mpc$", url):
                        result = get_test_data(url, "mpc", {"devices": []})
                    case url if re.match(r".*/rts/.*", url):
                        result = get_test_data(url, "rts", {"statistics": []})
                    case url if re.match(r".*/ambisense/.*/capability", url):
                        result = get_test_data(url, "rbr", {"rbrCapable": True})
                    case url if re.match(r".*/ambisense/.*/rooms", url):
                        result = get_test_data(url, "rooms", [])
                return result

            def unmatched_url(url, **kwargs):
                """
                Return test data based on the system ID in the URL
                """
                raise Exception(
                    f"Unmatched URL {url} with test data {self.test_data['_directory']}"
                )

            if self.test_data:
                # Create endpoints with stored JSON test data
                self.get(
                    re.compile(r".*/homes$"),
                    status=200,
                    payload=self.test_data["homes"],
                    repeat=True,
                )
                system_ids = [h["systemId"] for h in self.test_data["homes"]]
                if any(system_ids):
                    # Handle URLs that contain a system ID
                    self.get(
                        re.compile(rf".*({'|'.join(system_ids)}).*"),
                        callback=test_data_by_system,
                        repeat=True,
                    )
            self.get(
                re.compile(r".*"),
                callback=unmatched_url,
                repeat=True,
            )
            self.patch(
                re.compile(r".*"),
                callback=unmatched_url,
                repeat=True,
            )
            self.post(
                re.compile(r".*"),
                callback=unmatched_url,
                repeat=True,
            )
            self.delete(
                re.compile(r".*"),
                callback=unmatched_url,
                repeat=True,
            )
            return self

    return _mypyllant_aioresponses


async def _mocked_api(*args, **kwargs) -> MyPyllantAPI:
    api = MyPyllantAPI("test@example.com", "test", "vaillant", "germany")
    api.oauth_session = {
        "access_token": "access_token",
        "refresh_token": "refresh_token",
        "expires_in": 3600,
    }
    api.set_session_expires()
    return api


def list_test_data():
    test_data = []
    for d in [d for d in DATA_DIR.iterdir() if d.is_dir()]:
        # check if a json file exists in dir
        if list(d.rglob("*.json")):
            test_data.append(load_test_data(d))
    for f in DATA_DIR.glob("*.yaml"):
        test_data.append(load_test_data(f))
    return test_data


def load_test_data(data_dir: Path):
    if data_dir.is_dir():
        user_data: dict[str, Any] = defaultdict(dict)
        user_data["_directory"] = str(data_dir)
        for f in data_dir.rglob("*.json"):
            if f.parent != data_dir:
                user_data[f.parent.stem][f.stem] = json.loads(f.read_text())
            else:
                user_data[f.stem] = json.loads(f.read_text())
        return user_data
    else:
        return {"_directory": str(data_dir), **yaml.safe_load(data_dir.read_text())}



import pytest

from myPyllant.api import MyPyllantAPI
from myPyllant.tests.utils import _mocked_api, _mypyllant_aioresponses


@pytest.fixture
def mypyllant_aioresponses():
    return _mypyllant_aioresponses()


@pytest.fixture
async def mocked_api() -> MyPyllantAPI:
    return await _mocked_api()



#!/usr/bin/env python3

import argparse
import asyncio
import logging
from datetime import datetime, timedelta

from myPyllant.api import MyPyllantAPI
from myPyllant.const import ALL_COUNTRIES, BRANDS, DEFAULT_BRAND

parser = argparse.ArgumentParser(description="Export data from myVaillant API   .")
parser.add_argument("user", help="Username (email address) for the myVaillant app")
parser.add_argument("password", help="Password for the myVaillant app")
parser.add_argument(
    "brand",
    help="Brand your account is registered in, i.e. 'vaillant'",
    default=DEFAULT_BRAND,
    choices=BRANDS.keys(),
)
parser.add_argument(
    "--country",
    help="Country your account is registered in, i.e. 'germany'",
    choices=ALL_COUNTRIES.keys(),
    required=False,
)
parser.add_argument(
    "-v", "--verbose", help="increase output verbosity", action="store_true"
)


async def main(user, password, brand, country):
    async with MyPyllantAPI(user, password, brand, country) as api:
        async for system in api.get_systems():
            print(await api.set_set_back_temperature(system.zones[0], 18))
            print(await api.quick_veto_zone_temperature(system.zones[0], 21, 5))
            print(await api.cancel_quick_veto_zone_temperature(system.zones[0]))
            setpoint = 10.0 if system.control_identifier.is_vrc700 else None
            print(
                await api.set_holiday(
                    system,
                    datetime.now(system.timezone),
                    datetime.now(system.timezone) + timedelta(days=7),
                    setpoint,  # Setpoint is only required for VRC700 systems
                )
            )
            print(await api.cancel_holiday(system))
            if system.domestic_hot_water:
                print(await api.boost_domestic_hot_water(system.domestic_hot_water[0]))
                print(await api.cancel_hot_water_boost(system.domestic_hot_water[0]))
                print(
                    await api.set_domestic_hot_water_temperature(
                        system.domestic_hot_water[0], 46
                    )
                )


if __name__ == "__main__":
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    asyncio.run(main(args.user, args.password, args.brand, args.country))



#!/usr/bin/env python3
import re
from pathlib import Path

if __name__ == "__main__":
    readme = Path(__file__).resolve().parents[3] / "README.md"
    sample = Path(__file__).resolve().parents[1] / "sample.py"
    mypyllant_component = Path(__file__).resolve().parents[4] / "mypyllant-component"

    # README
    new_readme_content = f"""## Using the API in Python

```python
{sample.read_text()}
```
"""
    new_readme = re.sub(
        r"## Using the API in Python(.+?)```\n",
        new_readme_content,
        readme.read_text(),
        flags=re.DOTALL,
    )
    if readme.read_text() != new_readme:
        readme.write_text(
            re.sub(
                r"## Using the API in Python(.+?)```\n",
                new_readme_content,
                readme.read_text(),
                flags=re.DOTALL,
            )
        )
        print("README.md was updated with latest sample.py")

    # Docs
    if mypyllant_component.exists():
        docs = (
            Path(__file__).resolve().parents[4]
            / "mypyllant-component/docs/docs/3-library.md"
        )
        new_docs_content = f"""## Using the API in Python

```python
{sample.read_text()}
```
"""
        new_docs = re.sub(
            r"## Using the API in Python(.+?)```\n",
            new_docs_content,
            docs.read_text(),
            flags=re.DOTALL,
        )
        if docs.read_text() != new_docs:
            docs.write_text(
                re.sub(
                    r"## Using the API in Python(.+?)```\n",
                    new_docs_content,
                    docs.read_text(),
                    flags=re.DOTALL,
                )
            )
            print("docs/3-library.md was updated with latest sample.py")



from datetime import datetime, timezone

from myPyllant.utils import datetime_parse


async def test_datetime_parse() -> None:
    assert isinstance(
        datetime_parse("2022-03-28T19:37:12.27334Z", timezone.utc), datetime
    )
    assert isinstance(datetime_parse("2022-03-28T19:37:12Z", timezone.utc), datetime)



from __future__ import annotations

import calendar
import datetime
import logging
from collections.abc import Iterator
from dataclasses import asdict, fields, field
from typing import TypeVar, Any, Iterable

from pydantic import ConfigDict
from pydantic.dataclasses import dataclass

from myPyllant.const import BRANDS
from myPyllant.enums import (
    CircuitState,
    DeviceDataBucketResolution,
    ZoneOperatingMode,
    ZoneCurrentSpecialFunction,
    ZoneHeatingState,
    DHWCurrentSpecialFunction,
    DHWOperationMode,
    VentilationOperationMode,
    ControlIdentifier,
    ZoneOperatingModeVRC700,
    DHWCurrentSpecialFunctionVRC700,
    DHWOperationModeVRC700,
    AmbisenseRoomOperationMode,
    VentilationOperationModeVRC700,
    ZoneOperatingType,
    EnergyManagerState,
)
from myPyllant.utils import datetime_parse, prepare_field_value_for_dict

logger = logging.getLogger(__name__)

T = TypeVar("T", bound="MyPyllantDataClass")


config = ConfigDict(arbitrary_types_allowed=True)


@dataclass(kw_only=True, config=config)
class MyPyllantDataClass:
    """
    Base class that runs type validation in __init__ and can create an instance from API values
    """

    extra_fields: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_api(cls: type[T], **data) -> T:
        """
        Creates enums & dates from strings before calling __init__
        """
        dataclass_fields = fields(cls)
        extra_fields = set(data.keys()) - {f.name for f in dataclass_fields}
        datetime_fields = set(
            f.name
            for f in dataclass_fields
            if "datetime.datetime" in str(f.type)  # TODO: Use typing.get_origin()
        )
        timezone: datetime.tzinfo | None = data.get("timezone")

        if timezone is None and datetime_fields:
            raise ValueError(
                f"timezone is required in {cls.__name__}.from_api() for datetime field {', '.join(datetime_fields)}"
            )

        for k, v in data.items():
            if v is not None and k in datetime_fields and timezone is not None:
                if v.endswith("Z"):
                    # Some dates are returned as "2024-01-01T00:00:00Z" without timezone information
                    data[k] = datetime_parse(v, timezone)
                else:
                    # ... and some are ISO formatted with timezone information
                    data[k] = datetime.datetime.fromisoformat(v)

        if extra_fields:
            data["extra_fields"] = {f: data[f] for f in extra_fields}

        return cls(**data)

    def prepare_dict(self) -> dict:
        data = asdict(self)
        return prepare_field_value_for_dict(data)


@dataclass(config=config)
class Home(MyPyllantDataClass):
    country_code: str
    timezone: datetime.tzinfo
    nomenclature: str
    serial_number: str
    state: str
    system_id: str
    home_name: str | None = None
    address: dict = field(default_factory=dict)
    product_information: str | None = None
    migration_state: str | None = None
    cag: bool | None = None
    firmware_version: str | None = None
    firmware: dict = field(default_factory=dict)
    product_metadata: dict = field(default_factory=dict)

    @property
    def name(self):
        return f"{self.home_name} {self.nomenclature}"


@dataclass(config=config)
class BaseTimeProgramDay(MyPyllantDataClass):
    index: int
    weekday_name: str
    start_time: int
    end_time: int

    def __eq__(self, other):
        raise NotImplementedError

    @property
    def start_datetime_time(self) -> datetime.time:
        return datetime.time(self.start_time // 60, self.start_time % 60)

    @property
    def end_datetime_time(self) -> datetime.time:
        end_time = self.end_time % 1440
        return datetime.time(end_time // 60, end_time % 60)

    def start_datetime(self, date) -> datetime.datetime:
        return date.replace(
            hour=self.start_time // 60,
            minute=self.start_time % 60,
            second=0,
            microsecond=0,
        )

    def end_datetime(self, date) -> datetime.datetime:
        """
        end_time can be > 1440 for RoomTimeProgramDay, which indicates the time slot ends on a later day

        On all other time programs, end_time is <= 1440. Exactly 1440 is returned as midnight on the next day.
        """
        days = self.end_time // 1440
        end_time = self.end_time % 1440
        return date.replace(
            hour=end_time // 60,
            minute=end_time % 60,
            second=0,
            microsecond=0,
        ) + datetime.timedelta(days=days)


@dataclass(config=config)
class BaseTimeProgram(MyPyllantDataClass):
    monday: list
    tuesday: list
    wednesday: list
    thursday: list
    friday: list
    saturday: list
    sunday: list
    meta_info: dict | None = None

    @property
    def has_time_program(self):
        return any(
            [len(getattr(self, weekday)) > 0 for weekday in self.weekday_names()]
        )

    @classmethod
    def weekday_names(cls):
        return [w.lower() for w in calendar.day_name]

    def matching_weekdays(self, time_program_day: BaseTimeProgramDay):
        """
        Returns a list of weekday names that have a matching BaseTimeProgramDay
        """
        return [
            w
            for w in self.weekday_names()
            if any([d == time_program_day for d in getattr(self, w)])
        ]

    def check_overlap(self):
        for weekday in self.weekday_names():
            day_list: list[BaseTimeProgramDay] = getattr(self, weekday)
            if len(day_list) == 0:
                continue
            day_list.sort(key=lambda x: x.start_time)
            # Create non-overlapping BaseTimeProgramDays
            for i, day in enumerate(day_list):
                other_slots = day_list[i + 1 :]
                for other_slot in other_slots:
                    if (
                        day.end_time > other_slot.start_time
                        and day.start_time < other_slot.end_time
                    ):
                        raise ValueError(
                            f"Time program {day.start_datetime_time} - {day.end_datetime_time} "
                            f"overlaps with {other_slot.start_datetime_time} - {other_slot.end_datetime_time}"
                        )

    def as_datetime(
        self, start, end
    ) -> Iterator[tuple[ZoneTimeProgramDay, datetime.datetime, datetime.datetime]]:
        current = start
        while current < end:
            weekday = current.strftime("%A").lower()
            for time_program in getattr(self, weekday):
                if time_program.start_datetime(current) < end:
                    yield (
                        time_program,
                        time_program.start_datetime(current),
                        time_program.end_datetime(current),
                    )
                else:
                    break
            current += datetime.timedelta(days=1)

    @classmethod
    def create_day_from_api(cls, **kwargs):
        raise NotImplementedError

    @classmethod
    def from_api(cls, **data):
        for weekday_name in [w for w in cls.weekday_names()]:
            data[weekday_name] = [
                cls.create_day_from_api(index=i, weekday_name=weekday_name, **d)
                for i, d in enumerate(data.get(weekday_name, []))
            ]
        return super().from_api(**data)


@dataclass(config=config)
class ZoneTimeProgramDay(BaseTimeProgramDay):
    setpoint: float | None = None

    def __eq__(self, other):
        """
        When comparing two ZoneTimeProgramDay, we only care about start_time, end_time, and setpoint
        """
        if not isinstance(other, ZoneTimeProgramDay):
            return False
        return (
            self.start_time == other.start_time
            and self.end_time == other.end_time
            and self.setpoint == other.setpoint
        )


@dataclass(config=config)
class ZoneTimeProgram(BaseTimeProgram):
    monday: list[ZoneTimeProgramDay]
    tuesday: list[ZoneTimeProgramDay]
    wednesday: list[ZoneTimeProgramDay]
    thursday: list[ZoneTimeProgramDay]
    friday: list[ZoneTimeProgramDay]
    saturday: list[ZoneTimeProgramDay]
    sunday: list[ZoneTimeProgramDay]

    @classmethod
    def create_day_from_api(cls, **kwargs):
        return ZoneTimeProgramDay(**kwargs)

    def set_setpoint(
        self, temperature: float, update_similar_to_dow: str | None = None
    ):
        """
        Sets the setpoint on all weekdays of a time program to the new value
        """
        weekday_names = [w.lower() for w in calendar.day_name]
        if update_similar_to_dow and update_similar_to_dow not in weekday_names:
            raise ValueError(
                "%s is not a valid weekday, use one of %s or None",
                update_similar_to_dow,
                ", ".join(weekday_names),
            )
        # TODO: Implement update_similar_to_dow check
        for w in weekday_names:
            day_list: list[ZoneTimeProgramDay] = getattr(self, w)
            for d in day_list:
                d.setpoint = temperature


@dataclass(config=config)
class ZoneHeating(MyPyllantDataClass):
    control_identifier: ControlIdentifier
    operation_mode_heating: ZoneOperatingMode | ZoneOperatingModeVRC700
    set_back_temperature: float
    time_program_heating: ZoneTimeProgram | None = None
    manual_mode_setpoint_heating: float | None = None
    day_temperature_heating: float | None = None  # VRC700 only

    @classmethod
    def from_api(cls, **data):
        if "time_program_heating" in data:
            data["time_program_heating"] = ZoneTimeProgram.from_api(
                **data["time_program_heating"]
            )
        control_identifier: ControlIdentifier = data["control_identifier"]
        if control_identifier.is_vrc700:
            data["operation_mode_heating"] = ZoneOperatingModeVRC700(
                data["operation_mode_heating"]
            )
        else:
            data["operation_mode_heating"] = ZoneOperatingMode(
                data["operation_mode_heating"]
            )

        return super().from_api(**data)


@dataclass(config=config)
class ZoneCooling(MyPyllantDataClass):
    control_identifier: ControlIdentifier
    setpoint_cooling: float
    operation_mode_cooling: ZoneOperatingMode | ZoneOperatingModeVRC700
    time_program_cooling: ZoneTimeProgram | None = None
    manual_mode_setpoint_cooling: float | None = None

    @classmethod
    def from_api(cls, **data):
        if "time_program_cooling" in data:
            data["time_program_cooling"] = ZoneTimeProgram.from_api(
                **data["time_program_cooling"]
            )
        control_identifier: ControlIdentifier = data["control_identifier"]
        if control_identifier.is_vrc700:
            data["operation_mode_cooling"] = ZoneOperatingModeVRC700(
                data["operation_mode_cooling"]
            )
        else:
            data["operation_mode_cooling"] = ZoneOperatingMode(
                data["operation_mode_cooling"]
            )

        return super().from_api(**data)


@dataclass(config=config)
class ZoneGeneral(MyPyllantDataClass):
    name: str
    timezone: datetime.tzinfo
    holiday_start_date_time: datetime.datetime | None = None
    holiday_end_date_time: datetime.datetime | None = None

    @property
    def holiday_planned(self) -> bool:
        return (
            self.holiday_start_date_time is not None
            and self.holiday_end_date_time is not None
            and self.holiday_end_date_time > datetime.datetime.now(self.timezone)
        )

    @property
    def holiday_start_in_future(self) -> bool:
        return (
            self.holiday_start_date_time is not None
            and self.holiday_start_date_time > datetime.datetime.now(self.timezone)
        )

    @property
    def holiday_ongoing(self) -> bool:
        return (
            self.holiday_start_date_time is not None
            and self.holiday_end_date_time is not None
            and self.holiday_start_date_time
            < datetime.datetime.now(self.timezone)
            < self.holiday_end_date_time
        )

    @property
    def holiday_remaining(self) -> datetime.timedelta | None:
        return (
            self.holiday_end_date_time - datetime.datetime.now(self.timezone)
            if self.holiday_end_date_time and self.holiday_ongoing
            else None
        )


@dataclass(config=config)
class Circuit(MyPyllantDataClass):
    system_id: str
    index: int
    circuit_state: CircuitState | None = None
    mixer_circuit_type_external: str | None = None
    set_back_mode_enabled: bool | None = None
    zones: list = field(default_factory=list)
    is_cooling_allowed: bool | None = None
    current_circuit_flow_temperature: float | None = None
    heating_curve: float | None = None
    heating_flow_temperature_minimum_setpoint: float | None = None
    heating_flow_temperature_maximum_setpoint: float | None = None
    heat_demand_limited_by_outside_temperature: float | None = None
    min_flow_temperature_setpoint: float | None = None
    calculated_energy_manager_state: str | None = None


@dataclass(config=config)
class Zone(MyPyllantDataClass):
    system_id: str
    general: ZoneGeneral
    timezone: datetime.tzinfo
    control_identifier: ControlIdentifier
    index: int
    zone_binding: str
    heating: ZoneHeating
    current_special_function: ZoneCurrentSpecialFunction
    is_active: bool | None = None
    heating_state: ZoneHeatingState | None = None
    is_cooling_allowed: bool | None = None
    is_manual_cooling_active: bool | None = None
    cooling: ZoneCooling | None = None
    current_room_temperature: float | None = None
    desired_room_temperature_setpoint_heating: float | None = None
    desired_room_temperature_setpoint_cooling: float | None = None
    desired_room_temperature_setpoint: float | None = None
    current_room_humidity: float | None = None
    associated_circuit: Circuit | None = None
    associated_circuit_index: int | None = None
    quick_veto_start_date_time: datetime.datetime | None = None
    quick_veto_end_date_time: datetime.datetime | None = None

    @classmethod
    def from_api(cls, circuits: list[Circuit] | None = None, **data):
        data["heating"] = ZoneHeating.from_api(
            control_identifier=data["control_identifier"], **data["heating"]
        )
        if "cooling" in data:
            if not data["cooling"]:
                # cooling might be set as an empty dict
                data["cooling"] = None
            else:
                data["cooling"] = ZoneCooling.from_api(
                    control_identifier=data["control_identifier"], **data["cooling"]
                )
        data["general"] = ZoneGeneral.from_api(
            timezone=data["timezone"], **data["general"]
        )
        if circuits:
            data["associated_circuit"] = next(
                c for c in circuits if c.index == data["associated_circuit_index"]
            )
        return super().from_api(**data)

    @property
    def is_cooling_allowed_circuit(self):
        """
        On VRC700, is_cooling_allowed is only set on the circuit
        """
        return self.is_cooling_allowed or (
            self.associated_circuit and self.associated_circuit.is_cooling_allowed
        )

    @property
    def name(self):
        return self.general.name

    @property
    def quick_veto_ongoing(self) -> bool:
        return (
            self.quick_veto_start_date_time is not None
            and self.quick_veto_end_date_time is not None
            and self.quick_veto_start_date_time
            < datetime.datetime.now(self.timezone)
            < self.quick_veto_end_date_time
        )

    @property
    def quick_veto_remaining(self) -> datetime.timedelta | None:
        return (
            self.quick_veto_end_date_time - datetime.datetime.now(self.timezone)
            if self.quick_veto_end_date_time and self.quick_veto_ongoing
            else None
        )

    @property
    def is_eco_mode(self) -> bool:
        return (
            self.is_auto_heating_mode
            and self.current_special_function == ZoneCurrentSpecialFunction.NONE
            and self.desired_room_temperature_setpoint == 0.0
        )

    @property
    def is_auto_heating_mode(self) -> bool:
        """
        Whether the zone is in time-controlled heating mode
        """
        return self.heating.operation_mode_heating in [
            ZoneOperatingMode.TIME_CONTROLLED,
            ZoneOperatingModeVRC700.AUTO,
        ]

    @property
    def active_operating_type(self) -> str:
        """
        Whether the system is cooling or heating, based on the desired temperature
        """
        if (
            self.desired_room_temperature_setpoint
            == self.desired_room_temperature_setpoint_cooling
        ):
            return ZoneOperatingType.COOLING
        else:
            return ZoneOperatingType.HEATING

    @property
    def active_operation_mode(
        self,
    ) -> ZoneOperatingMode | ZoneOperatingModeVRC700 | None:
        """
        Returns the active operation mode, of the active operating type
        """
        operation: ZoneHeating | ZoneCooling | None = getattr(
            self, self.active_operating_type
        )
        if operation:
            return getattr(operation, f"operation_mode_{self.active_operating_type}")
        else:
            return None


@dataclass(config=config)
class DHWTimeProgramDay(BaseTimeProgramDay):
    def __eq__(self, other):
        """
        When comparing two DHWTimeProgramDay, we only care about start_time and end_time
        """
        if not isinstance(other, DHWTimeProgramDay):
            return False
        return self.start_time == other.start_time and self.end_time == other.end_time


@dataclass(config=config)
class DHWTimeProgram(BaseTimeProgram):
    monday: list[DHWTimeProgramDay]
    tuesday: list[DHWTimeProgramDay]
    wednesday: list[DHWTimeProgramDay]
    thursday: list[DHWTimeProgramDay]
    friday: list[DHWTimeProgramDay]
    saturday: list[DHWTimeProgramDay]
    sunday: list[DHWTimeProgramDay]

    @classmethod
    def create_day_from_api(cls, **kwargs):
        return DHWTimeProgramDay(**kwargs)


@dataclass(config=config)
class DomesticHotWater(MyPyllantDataClass):
    system_id: str
    index: int
    control_identifier: ControlIdentifier
    current_special_function: (
        DHWCurrentSpecialFunction | DHWCurrentSpecialFunctionVRC700
    )
    max_setpoint: float
    min_setpoint: float
    operation_mode_dhw: DHWOperationMode | DHWOperationModeVRC700
    time_program_dhw: DHWTimeProgram
    time_program_circulation_pump: DHWTimeProgram
    current_dhw_temperature: float | None = None
    tapping_setpoint: float | None = None

    @property
    def is_cylinder_boosting(self) -> bool:
        return self.current_special_function in [
            DHWCurrentSpecialFunction.CYLINDER_BOOST,
            DHWCurrentSpecialFunctionVRC700.CYLINDER_BOOST,
        ]

    @classmethod
    def from_api(cls, **data):
        data["time_program_dhw"] = DHWTimeProgram.from_api(**data["time_program_dhw"])
        data["time_program_circulation_pump"] = DHWTimeProgram.from_api(
            **data["time_program_circulation_pump"]
        )
        control_identifier: ControlIdentifier = data["control_identifier"]
        if control_identifier.is_vrc700:
            data["current_special_function"] = DHWCurrentSpecialFunctionVRC700(
                data["current_special_function"]
            )
            data["operation_mode_dhw"] = DHWOperationModeVRC700(
                data["operation_mode_dhw"]
            )
        else:
            data["current_special_function"] = DHWCurrentSpecialFunction(
                data["current_special_function"]
            )
            data["operation_mode_dhw"] = DHWOperationMode(data["operation_mode_dhw"])
        return super().from_api(**data)


@dataclass(config=config)
class Ventilation(MyPyllantDataClass):
    system_id: str
    index: int
    control_identifier: ControlIdentifier
    maximum_day_fan_stage: int
    maximum_night_fan_stage: int
    operation_mode_ventilation: (
        VentilationOperationMode | VentilationOperationModeVRC700
    )
    time_program_ventilation: dict

    @classmethod
    def from_api(cls, **data):
        control_identifier: ControlIdentifier = data["control_identifier"]
        if control_identifier.is_vrc700:
            data["operation_mode_ventilation"] = VentilationOperationModeVRC700(
                data["operation_mode_ventilation"]
            )
        else:
            data["operation_mode_ventilation"] = VentilationOperationMode(
                data["operation_mode_ventilation"]
            )

        return super().from_api(**data)


@dataclass(config=config)
class DeviceDataBucket(MyPyllantDataClass):
    start_date: datetime.datetime
    end_date: datetime.datetime
    value: float | None = None


@dataclass(config=config)
class DeviceData(MyPyllantDataClass):
    operation_mode: str
    skip_data_update: bool = False
    device: Any = None
    data_from: datetime.datetime | None = None
    data_to: datetime.datetime | None = None
    start_date: datetime.datetime | None = None
    end_date: datetime.datetime | None = None
    resolution: DeviceDataBucketResolution | None = None
    energy_type: str | None = None
    value_type: str | None = None
    calculated: bool | None = None
    total_consumption: float | None = None
    data: list[DeviceDataBucket] = field(default_factory=list)

    @property
    def total_consumption_rounded(self) -> float:
        """
        Rounds odd float values from the API to match the app, i.e. 8998.87 -> 9000 and 8499.99 -> 8500
        """
        return (
            round(self.total_consumption / 1000, 1) * 1000
            if self.total_consumption
            else 0.0
        )

    @classmethod
    def from_api(cls, **data):
        data["data_from"] = data.pop("from", None)
        data["data_to"] = data.pop("to", None)
        data["data"] = [
            DeviceDataBucket.from_api(timezone=data["timezone"], **dd)
            for dd in data.pop("data", [])
        ]
        return super().from_api(**data)


@dataclass(config=config)
class Device(MyPyllantDataClass):
    system_id: str
    timezone: datetime.tzinfo
    device_uuid: str
    ebus_id: str
    article_number: str
    device_serial_number: str
    type: str
    device_type: str
    first_data: datetime.datetime
    last_data: datetime.datetime
    brand: str
    name: str | None = None
    product_name: str | None = None
    spn: int | None = None
    bus_coupler_address: int | None = None
    emf_valid: bool | None = None
    operational_data: dict = field(default_factory=dict)
    data: list[DeviceData] = field(default_factory=list)
    properties: list = field(default_factory=list)
    mpc: dict | None = None
    rts_statistics: dict | None = None
    diagnostic_trouble_codes: list | None = None

    @property
    def name_display(self) -> str:
        """
        Product name might be empty, fall back to title-cased device type
        """
        return self.name or self.product_name_display

    @property
    def product_name_display(self) -> str:
        """
        Product name might be None, fall back to title-cased device type
        """
        if self.product_name:
            return (
                self.product_name.title()
                if self.product_name.islower()
                else self.product_name
            )
        else:
            return self.device_type.replace("_", "").title()

    @property
    def brand_name(self) -> str:
        return BRANDS[self.brand]

    @property
    def on_off_cycles(self) -> int | None:
        return self.rts_statistics.get("on_off_cycles") if self.rts_statistics else None

    @property
    def operation_time(self) -> int | None:
        return (
            self.rts_statistics.get("operation_time") if self.rts_statistics else None
        )

    @property
    def current_power(self) -> int | None:
        return self.mpc.get("current_power") if self.mpc else None

    @classmethod
    def from_api(cls, **data):
        device_data = []
        if "data" in data:
            for dd in data["data"]:
                if "timezone" not in dd:
                    dd["timezone"] = data["timezone"]
                device_data.append(DeviceData.from_api(**dd))
        data["data"] = device_data
        return super().from_api(**data)


@dataclass(config=config)
class RoomTimeProgramDay(BaseTimeProgramDay):
    temperature_setpoint: float | None = None

    @property
    def start_datetime_time(self) -> datetime.time:
        return datetime.time(self.start_time // 60, self.start_time % 60)

    def start_datetime(self, date) -> datetime.datetime:
        return date.replace(
            hour=self.start_time // 60,
            minute=self.start_time % 60,
            second=0,
            microsecond=0,
        )

    def __eq__(self, other):
        """
        When comparing two ZoneTimeProgramDay, we only care about start_time, end_time, and setpoint
        """
        if not isinstance(other, RoomTimeProgramDay):
            return False
        return (
            self.start_time == other.start_time
            and self.temperature_setpoint == other.temperature_setpoint
        )


@dataclass(config=config)
class RoomTimeProgram(BaseTimeProgram):
    monday: list[RoomTimeProgramDay]
    tuesday: list[RoomTimeProgramDay]
    wednesday: list[RoomTimeProgramDay]
    thursday: list[RoomTimeProgramDay]
    friday: list[RoomTimeProgramDay]
    saturday: list[RoomTimeProgramDay]
    sunday: list[RoomTimeProgramDay]

    @classmethod
    def dict_factory(cls, obj: Iterable[tuple[str, Any]]) -> dict[str, Any]:
        """
        Only includes certain fields when converting to dict with asdict()
        See https://stackoverflow.com/a/76017464

        Example::

            asdict(time_program, dict_factory=RoomTimeProgram.dict_factory)
        """
        include_fields = cls.weekday_names() + ["temperature_setpoint", "start_time"]
        return {k: v for (k, v) in obj if ((v is not None) and (k in include_fields))}

    @classmethod
    def from_api(cls, **data) -> RoomTimeProgram:
        """
        Unlike the other time programs, the Ambisense room time program does not have an end_time.
        Instead, the end_time is the start_time of the next slot.

        It takes this: {"monday": [{"start_time": 360, "temperature_setpoint": 21.0}, {"start_time": 480, "temperature_setpoint": 22.0}]}
        ...and converts it to this: {"monday": [RoomTimeProgramDay(start_time=360, end_time=480, temperature_setpoint=21.0), RoomTimeProgramDay(start_time=480, end_time=10440, temperature_setpoint=22.0)]}

        In RoomTimeProgramDay, end_time can be greater than 1440 (24h), if it ends on a later day.
        """

        # Keep a copy of the original data, because we need to traverse it to find the next time slot,
        # but we want to overwrite the weekdays with RoomTimeProgramDay instances
        orig_data = data.copy()
        for weekday_index, weekday_name in enumerate(cls.weekday_names()):
            weekday_slots = data.pop(weekday_name, [])
            # Make sure slots are sorted ascending by start time, so the next slot is always later
            sorted(weekday_slots, key=lambda x: x["start_time"])
            data[weekday_name] = []
            for slot_index, weekday_slot in enumerate(weekday_slots):
                if "end_time" in weekday_slot:
                    raise ValueError(
                        "Ambisense room time program does not allow setting end_time"
                    )
                if len(weekday_slots) > slot_index + 1:
                    # If there is another slot on the same day, use its start_time as the end_time
                    end_time = weekday_slots[slot_index + 1]["start_time"]
                else:
                    end_time = None
                    i = 0
                    # Iterate over weekdays until we find one with a time program slot
                    while not end_time:
                        i += 1
                        next_weekday_index = (weekday_index + i) % 7
                        next_day = orig_data.get(
                            cls.weekday_names()[next_weekday_index], []
                        )
                        if next_day:
                            # 24h == 1440min
                            end_time = (i * 1440) + next_day[0]["start_time"]
                        elif i == 7:
                            raise ValueError(
                                "Could not find end_time for time program %s"
                                % weekday_slot
                            )

                data[weekday_name].append(
                    cls.create_day_from_api(
                        index=slot_index,
                        end_time=end_time,
                        weekday_name=weekday_name,
                        **weekday_slot,
                    )
                )
        # Skip from_api of BaseTimeProgram, because we already did the conversion to TimeProgramDay
        return super(BaseTimeProgram, cls).from_api(**data)

    @classmethod
    def create_day_from_api(cls, **kwargs):
        if "setpoint" in kwargs:
            # Allow setpoint as well as temperature_setpoint, for consistency with the other classes
            kwargs["temperature_setpoint"] = kwargs.pop("setpoint")
        return RoomTimeProgramDay(**kwargs)


@dataclass(config=config)
class AmbisenseDevice(MyPyllantDataClass):
    device_type: str
    name: str
    sgtin: str
    unreach: bool | None = None
    low_bat: bool | None = None
    rssi: int | None = None
    rssi_peer: int | None = None


@dataclass(config=config)
class AmbisenseRoomConfiguration(MyPyllantDataClass):
    name: str
    timezone: datetime.tzinfo
    operation_mode: AmbisenseRoomOperationMode | None = None
    current_temperature: float | None = None
    temperature_setpoint: float | None = None
    icon_id: str | None = None
    current_humidity: float | None = None
    button_lock: bool | None = None
    window_state: bool | None = None
    quick_veto_end_time: datetime.datetime | None = None
    devices: list[AmbisenseDevice] = field(default_factory=list)

    @classmethod
    def from_api(cls, **data):
        if data["operation_mode"]:
            data["operation_mode"] = AmbisenseRoomOperationMode(
                data["operation_mode"].upper()
            )
        return super().from_api(**data)


@dataclass(config=config)
class AmbisenseRoom(MyPyllantDataClass):
    system_id: str
    room_index: int
    room_configuration: AmbisenseRoomConfiguration
    time_program: RoomTimeProgram

    @property
    def name(self) -> str:
        return self.room_configuration.name

    @classmethod
    def from_api(cls, **data):
        timezone = data.pop("timezone")
        data["time_program"] = RoomTimeProgram.from_api(**data["time_program"])
        data["room_configuration"] = AmbisenseRoomConfiguration.from_api(
            timezone=timezone, **data["room_configuration"]
        )
        return super().from_api(**data)


@dataclass(config=config)
class System(MyPyllantDataClass):
    id: str
    state: dict
    configuration: dict
    home: Home
    brand: str
    timezone: datetime.tzinfo
    control_identifier: ControlIdentifier
    connected: bool | None = None
    diagnostic_trouble_codes: list | None = None
    properties: dict = field(default_factory=dict)
    current_system: dict = field(default_factory=dict)
    zones: list[Zone] = field(default_factory=list)
    circuits: list[Circuit] = field(default_factory=list)
    domestic_hot_water: list[DomesticHotWater] = field(default_factory=list)
    ventilation: list[Ventilation] = field(default_factory=list)
    devices: list[Device] = field(default_factory=list)
    mpc: dict | None = None
    rts: dict | None = None
    energy_management: dict | None = None
    eebus: dict | None = None
    ambisense_capability: bool = False
    ambisense_rooms: list[AmbisenseRoom] = field(default_factory=list)

    @classmethod
    def from_api(cls, **data):
        if "home" in data and "id" not in data:
            data["id"] = data["home"].system_id
        ambisense_rooms = data.pop("ambisense_rooms")
        system: System = super().from_api(**data)
        logger.debug(f"Creating related models from state: {data}")
        system.extra_fields = system.merge_extra_fields()
        system.circuits = [
            Circuit.from_api(system_id=system.id, timezone=system.timezone, **c)
            for c in system.merge_object("circuits")
        ]
        system.zones = [
            Zone.from_api(
                system_id=system.id,
                timezone=system.timezone,
                control_identifier=system.control_identifier,
                circuits=system.circuits,
                **z,
            )
            for z in system.merge_object("zones")
        ]
        system.domestic_hot_water = [
            DomesticHotWater.from_api(
                system_id=system.id,
                timezone=system.timezone,
                control_identifier=system.control_identifier,
                **d,
            )
            for d in system.merge_object("dhw")
        ]
        # TODO: Is it called ventilations everywhere, or just on VRC700 controllers?
        if "ventilations" in system.configuration:
            ventilation_key = "ventilations"
        else:
            ventilation_key = "ventilation"
        system.ventilation = [
            Ventilation.from_api(
                system_id=system.id,
                control_identifier=system.control_identifier,
                timezone=system.timezone,
                **d,
            )
            for d in system.merge_object(ventilation_key)
        ]
        system.devices = [
            Device.from_api(
                system_id=system.id,
                timezone=system.timezone,
                type=k,
                brand=system.brand,
                **v,
            )
            for k, v in system.raw_devices
        ]
        system.ambisense_rooms = [
            AmbisenseRoom.from_api(system_id=system.id, timezone=system.timezone, **r)
            for r in ambisense_rooms
        ]
        return system

    def apply_diagnostic(self, device):
        dtc = self.diagnostic_trouble_codes_by_serial_number(
            device["device_serial_number"]
        )
        device["diagnostic_trouble_codes"] = dtc

    def apply_rts(self, device):
        rts_statistics = self.rts_statistics_by_device_uuid(device["device_uuid"])
        if rts_statistics:
            device["rts_statistics"] = rts_statistics

    def apply_mpc(self, device):
        mpc = self.mpc_by_device_uuid(device["device_uuid"])
        if mpc:
            device["mpc"] = mpc

    @property
    def raw_devices(self) -> Iterator[tuple[str, dict]]:
        for key, device in self.current_system.items():
            if isinstance(device, list) and key == "secondary_heat_generators":
                for secdevice in device:
                    if isinstance(secdevice, dict) and "device_uuid" in secdevice:
                        self.apply_diagnostic(secdevice)
                        self.apply_rts(secdevice)
                        self.apply_mpc(secdevice)
                        yield key, secdevice
            if isinstance(device, dict) and "device_uuid" in device:
                self.apply_diagnostic(device)
                self.apply_rts(device)
                self.apply_mpc(device)
                yield key, device

    @property
    def primary_heat_generator(self) -> Device | None:
        devices = [d for d in self.devices if d.type == "primary_heat_generator"]
        if len(devices) > 0:
            return devices[0]
        return None

    def merge_extra_fields(self) -> dict:
        return (
            self.extra_fields
            | self.configuration.get("system", {})
            | self.state.get("system", {})
            | self.properties.get("system", {})
        )

    def merge_object(self, obj_name) -> Iterator[dict]:
        """
        The Vaillant API returns information about zones, circuits, and dhw separately as
        configuration, state, and properties.

        This function merges everything together into one big dict for a given object (i.e. zones)
        """
        indexes = [o["index"] for o in self.configuration.get(obj_name, [])]
        for idx in indexes:
            # State and properties get merged into configuration
            configuration = next(
                c for c in self.configuration.get(obj_name, []) if c["index"] == idx
            )
            try:
                state = next(
                    c for c in self.state.get(obj_name, []) if c["index"] == idx
                )
            except (StopIteration, KeyError) as e:
                logger.debug("Error when merging state", exc_info=e)
                state = {}
            try:
                properties = next(
                    c for c in self.properties.get(obj_name, []) if c["index"] == idx
                )
            except (StopIteration, KeyError) as e:
                logger.debug("Error when merging properties", exc_info=e)
                properties = {}
            configuration.update(state)
            configuration.update(properties)
            yield configuration

    @property
    def outdoor_temperature(self) -> float | None:
        try:
            return self.state["system"]["outdoor_temperature"]
        except KeyError:
            logger.debug(
                "Could not get outdoor temperature from system control state",
            )
            return None

    @property
    def outdoor_temperature_average_24h(self) -> float | None:
        try:
            return self.state["system"]["outdoor_temperature_average_24h"]
        except KeyError:
            logger.debug(
                "Could not get outdoor temperature average 24h from system control state",
            )
            return None

    @property
    def water_pressure(self) -> float | None:
        try:
            return self.state["system"]["system_water_pressure"]
        except KeyError:
            logger.debug("Could not get water pressure from system control state")
            return None

    @property
    def cylinder_temperature_sensor_top_dhw(self) -> float | None:
        try:
            return self.state["system"]["cylinder_temperature_sensor_top_d_h_w"]
        except KeyError:
            logger.debug(
                "Could not get top DHW cylinder temperature from system control state"
            )
            return None

    @property
    def cylinder_temperature_sensor_bottom_dhw(self) -> float | None:
        try:
            return self.state["system"]["cylinder_temperature_sensor_bottom_d_h_w"]
        except KeyError:
            logger.debug(
                "Could not get bottom DHW cylinder temperature from system control state"
            )
            return None

    @property
    def cylinder_temperature_sensor_top_ch(self) -> float | None:
        try:
            return self.state["system"]["cylinder_temperature_sensor_top_c_h"]
        except KeyError:
            logger.debug(
                "Could not get top CH cylinder temperature from system control state"
            )
            return None

    @property
    def cylinder_temperature_sensor_bottom_ch(self) -> float | None:
        try:
            return self.state["system"]["cylinder_temperature_sensor_bottom_c_h"]
        except KeyError:
            logger.debug(
                "Could not get bottom CH cylinder temperature from system control state"
            )
            return None

    @property
    def system_flow_temperature(self) -> float | None:
        try:
            return self.state["system"]["system_flow_temperature"]
        except KeyError:
            logger.debug(
                "Could not get system flow temperature from system control state"
            )
            return None

    @property
    def energy_manager_state(self) -> EnergyManagerState | None:
        try:
            return EnergyManagerState(self.state["system"]["energy_manager_state"])
        except (KeyError, ValueError) as e:
            logger.debug(
                "Could not get energy manager state from system control state",
                exc_info=e,
            )
            return None

    @property
    def is_cooling_allowed(self) -> bool:
        return any([z.is_cooling_allowed for z in self.zones]) or any(
            [c.is_cooling_allowed for c in self.circuits]
        )

    @property
    def manual_cooling_start_date(self) -> datetime.datetime | None:
        manual_cooling_start_date = self.configuration.get("system", {}).get(
            "manual_cooling_start_date"
        )
        if manual_cooling_start_date:
            return datetime_parse(
                manual_cooling_start_date,
                self.timezone,
            )
        return None

    @property
    def manual_cooling_end_date(self) -> datetime.datetime | None:
        manual_cooling_end_date = self.configuration.get("system", {}).get(
            "manual_cooling_end_date"
        )
        if manual_cooling_end_date:
            return datetime_parse(
                manual_cooling_end_date,
                self.timezone,
            )
        return None

    @property
    def manual_cooling_days(self) -> int | None:
        if self.control_identifier.is_vrc700:
            return self.configuration.get("system", {}).get("cooling_for_x_days")
        elif self.manual_cooling_end_date:
            return (
                self.manual_cooling_end_date - datetime.datetime.now(self.timezone)
            ).days
        else:
            return None

    @property
    def system_name(self) -> str:
        if self.primary_heat_generator:
            return self.primary_heat_generator.product_name_display
        elif self.home:
            return self.home.nomenclature
        else:
            return "System"

    @property
    def brand_name(self) -> str:
        return BRANDS[self.brand]

    @property
    def has_diagnostic_trouble_codes(self) -> bool:
        return self.diagnostic_trouble_codes is not None and any(
            [len(d["codes"]) > 0 for d in self.diagnostic_trouble_codes]
        )

    def diagnostic_trouble_codes_by_serial_number(
        self, serial_number: str
    ) -> list | None:
        if self.diagnostic_trouble_codes:
            dtc = [
                d
                for d in self.diagnostic_trouble_codes
                if d["serial_number"] == serial_number
            ]
            return dtc[0]["codes"] if dtc else None
        else:
            return None

    def rts_statistics_by_device_uuid(self, device_uuid: str) -> dict | None:
        if not self.rts:
            return None
        statistics = [
            s for s in self.rts.get("statistics", []) if s["device_id"] == device_uuid
        ]
        return statistics[0] if statistics else None

    def mpc_by_device_uuid(self, device_uuid: str) -> dict | None:
        if not self.mpc:
            return None
        mpc = [m for m in self.mpc.get("devices", []) if m["device_id"] == device_uuid]
        return mpc[0] if mpc else None

    @property
    def manual_cooling_planned(self) -> bool:
        return (
            self.manual_cooling_start_date is not None
            and self.manual_cooling_end_date is not None
            and self.manual_cooling_end_date > datetime.datetime.now(self.timezone)
        )

    @property
    def manual_cooling_start_in_future(self) -> bool:
        return (
            self.manual_cooling_start_date is not None
            and self.manual_cooling_start_date > datetime.datetime.now(self.timezone)
        )

    @property
    def manual_cooling_ongoing(self) -> bool:
        return (
            self.manual_cooling_start_date is not None
            and self.manual_cooling_end_date is not None
            and self.manual_cooling_start_date
            < datetime.datetime.now(self.timezone)
            < self.manual_cooling_end_date
        )

    @property
    def manual_cooling_remaining(self) -> datetime.timedelta | None:
        return (
            self.manual_cooling_end_date - datetime.datetime.now(self.timezone)
            if self.manual_cooling_end_date and self.manual_cooling_ongoing
            else None
        )


@dataclass(config=config)
class SystemReport(MyPyllantDataClass):
    file_name: str
    file_content: str



import datetime

import pytest
from pydantic import ValidationError

from myPyllant.api import MyPyllantAPI
from .generate_test_data import DATA_DIR

from ..models import (
    Home,
    System,
    ZoneHeating,
    ZoneTimeProgram,
    Device,
    AmbisenseRoom,
    RoomTimeProgram,
    AmbisenseDevice,
    Circuit,
)
from ..enums import ZoneOperatingMode, ControlIdentifier, ZoneOperatingType
from .utils import list_test_data, load_test_data


@pytest.mark.parametrize("test_data", list_test_data())
async def test_systems(
    mypyllant_aioresponses, mocked_api: MyPyllantAPI, test_data
) -> None:
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(mocked_api.get_systems())

        assert isinstance(system, System), "Expected System return type"
        assert isinstance(system.home, Home)
        assert isinstance(system.home.system_id, str)
        await mocked_api.aiohttp_session.close()


async def test_type_validation() -> None:
    with pytest.raises(ValidationError):
        ZoneHeating.from_api(
            control_identifier=ControlIdentifier.TLI,
            manual_mode_setpoint_heating="shouldbenumber",
            operation_mode_heating=ZoneOperatingMode.MANUAL,
            time_program_heating={},
            set_back_temperature=15.0,
        )
    with pytest.raises(ValueError):
        ZoneHeating.from_api(
            control_identifier=ControlIdentifier.TLI,
            manual_mode_setpoint_heating=15.0,
            operation_mode_heating="INVALID",
            time_program_heating={},
            set_back_temperature=15.0,
        )


@pytest.mark.parametrize("test_data", list_test_data())
async def test_trouble_codes(
    mypyllant_aioresponses, mocked_api: MyPyllantAPI, test_data
) -> None:
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(
            mocked_api.get_systems(include_diagnostic_trouble_codes=True)
        )
        assert isinstance(system.diagnostic_trouble_codes, list)
        if not system.diagnostic_trouble_codes:
            await mocked_api.aiohttp_session.close()
            pytest.skip("No diagnostic trouble codes in test data, skipping")
        dtc = system.diagnostic_trouble_codes[0]
        assert isinstance(dtc, dict)
        assert isinstance(dtc["codes"], list)

        system.diagnostic_trouble_codes[0]["codes"] = [1, 2]
        assert system.has_diagnostic_trouble_codes

        system.diagnostic_trouble_codes = [{"codes": []}]
        assert not system.has_diagnostic_trouble_codes
        await mocked_api.aiohttp_session.close()


@pytest.mark.parametrize("test_data", list_test_data())
async def test_circuit_association(
    mypyllant_aioresponses, mocked_api: MyPyllantAPI, test_data
) -> None:
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(
            mocked_api.get_systems(include_diagnostic_trouble_codes=True)
        )
        for zone in system.zones:
            assert isinstance(zone.associated_circuit, Circuit)
    await mocked_api.aiohttp_session.close()


async def test_operating_type(mypyllant_aioresponses, mocked_api: MyPyllantAPI) -> None:
    test_data = load_test_data(DATA_DIR / "ventilation")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(
            mocked_api.get_systems(include_diagnostic_trouble_codes=True)
        )
        assert system.zones[0].active_operating_type == ZoneOperatingType.HEATING
        assert system.zones[0].active_operation_mode == ZoneOperatingMode.MANUAL
    test_data = load_test_data(DATA_DIR / "heatpump_cooling")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(
            mocked_api.get_systems(include_diagnostic_trouble_codes=True)
        )
        assert system.zones[0].active_operating_type == ZoneOperatingType.COOLING
    await mocked_api.aiohttp_session.close()


async def test_cooling_allowed(
    mypyllant_aioresponses, mocked_api: MyPyllantAPI
) -> None:
    test_data = load_test_data(DATA_DIR / "vrc700")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(
            mocked_api.get_systems(include_diagnostic_trouble_codes=True)
        )
        assert system.zones[0].associated_circuit.is_cooling_allowed is False
        assert system.zones[0].is_cooling_allowed_circuit is False
    await mocked_api.aiohttp_session.close()


async def test_ventilation(mypyllant_aioresponses, mocked_api: MyPyllantAPI) -> None:
    data_list = [
        load_test_data(DATA_DIR / "vrc700_ventilation.yaml"),
        load_test_data(DATA_DIR / "ventilation"),
    ]
    for test_data in data_list:
        with mypyllant_aioresponses(test_data) as _:
            system = await anext(mocked_api.get_systems())
            assert isinstance(system.ventilation, list)
            assert len(system.ventilation) > 0
            devices = [d for d in system.devices if d.type == "ventilation"]
            assert devices[0].device_type == "VENTILATION"
    await mocked_api.aiohttp_session.close()


async def test_time_program_overlap() -> None:
    # Should not raise an exception
    time_program = ZoneTimeProgram.from_api(
        **{
            "monday": [
                {"start_time": 360, "end_time": 1320, "setpoint": 21.0},
                {"start_time": 1320, "end_time": 1340, "setpoint": 22.0},
            ]
        }
    )
    time_program.check_overlap()

    time_program = ZoneTimeProgram.from_api(
        **{
            "monday": [
                {"start_time": 360, "end_time": 1320, "setpoint": 21.0},
                {"start_time": 380, "end_time": 420, "setpoint": 22.0},
            ]
        }
    )
    with pytest.raises(ValueError):
        time_program.check_overlap()


async def test_rts_statistics(mypyllant_aioresponses, mocked_api: MyPyllantAPI) -> None:
    data_list = [
        load_test_data(DATA_DIR / "vrc700_mpc_rts.yaml"),
        load_test_data(DATA_DIR / "rts"),
    ]
    for test_data in data_list:
        with mypyllant_aioresponses(test_data) as _:
            system = await anext(mocked_api.get_systems(include_rts=True))
            assert isinstance(system.rts, dict)
            assert len(system.rts.get("statistics", [])) > 0
            rts_device_id = system.rts["statistics"][0]["device_id"]
            d: Device = [d for d in system.devices if d.device_uuid == rts_device_id][0]
            assert isinstance(d.on_off_cycles, int)
            assert isinstance(d.operation_time, int)
    await mocked_api.aiohttp_session.close()


async def test_mpc(mypyllant_aioresponses, mocked_api: MyPyllantAPI) -> None:
    test_data = load_test_data(DATA_DIR / "vrc700_mpc_rts.yaml")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(mocked_api.get_systems(include_mpc=True))
        assert isinstance(system.mpc, dict)
        assert len(system.mpc.get("devices", [])) > 0
        mpc_device_id = system.mpc["devices"][0]["device_id"]
        d: Device = [d for d in system.devices if d.device_uuid == mpc_device_id][0]
        assert isinstance(d.current_power, int)
    await mocked_api.aiohttp_session.close()


async def test_extra_system_state_properties(
    mypyllant_aioresponses, mocked_api: MyPyllantAPI
) -> None:
    test_data = load_test_data(DATA_DIR / "two_systems")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(mocked_api.get_systems())
        assert system.cylinder_temperature_sensor_top_ch is not None
        assert system.cylinder_temperature_sensor_top_dhw is not None
        assert system.cylinder_temperature_sensor_bottom_dhw is not None
    await mocked_api.aiohttp_session.close()


async def test_ambisense(mypyllant_aioresponses, mocked_api: MyPyllantAPI) -> None:
    test_data = load_test_data(DATA_DIR / "ambisense2.yaml")
    with mypyllant_aioresponses(test_data) as _:
        system = await anext(mocked_api.get_systems(include_ambisense_rooms=True))
        assert len(system.ambisense_rooms) > 0
        assert isinstance(system.ambisense_rooms[0], AmbisenseRoom)
        for room in system.ambisense_rooms:
            assert isinstance(room.name, str)
            assert isinstance(room.room_configuration.devices[0], AmbisenseDevice)
            assert len(room.room_configuration.devices[0].name) > 0
            assert isinstance(room.time_program, RoomTimeProgram)
            if room.room_configuration.operation_mode:
                assert isinstance(room.room_configuration.current_temperature, float)
            if room.room_index == 1:
                assert isinstance(room.time_program.monday[0].start_time, int)
                assert isinstance(
                    room.time_program.monday[0].temperature_setpoint, float
                )
    await mocked_api.aiohttp_session.close()


async def test_ambisense_time_program() -> None:
    time_program = RoomTimeProgram.from_api(
        **{
            "monday": [
                {"start_time": 360, "temperature_setpoint": 21.0},
                {"start_time": 480, "temperature_setpoint": 22.0},
            ],
            "tuesday": [
                {"start_time": 300, "temperature_setpoint": 20.0},
            ],
        }
    )
    assert time_program.monday[0].temperature_setpoint == 21.0
    assert time_program.monday[0].start_time == 360
    assert time_program.monday[0].end_time == 480
    test_date = datetime.datetime(2024, 1, 1, 0, 0, 0)  # Monday
    assert time_program.monday[0].end_datetime(test_date) == datetime.datetime(
        2024, 1, 1, 8, 0
    ), "Expected end time to be 8:00 on the same day"

    assert time_program.monday[1].temperature_setpoint == 22.0
    assert time_program.monday[1].start_time == 480
    assert time_program.monday[1].end_time == 1740
    assert time_program.monday[1].end_datetime(test_date) == datetime.datetime(
        2024, 1, 2, 5, 0
    ), "Expected end time to be 5:00 on the next day"

    test_date = datetime.datetime(2024, 1, 2, 0, 0, 0)  # Tuesday
    assert time_program.tuesday[0].temperature_setpoint == 20.0
    assert time_program.tuesday[0].start_time == 300
    assert time_program.tuesday[0].end_time == 360 + 1440 * 6
    assert time_program.tuesday[0].end_datetime(test_date) == datetime.datetime(
        2024, 1, 8, 6, 0
    ), "Expected end time to be Monday 6:00 one week later"

    time_program = RoomTimeProgram.from_api(
        **{
            "monday": [
                {"start_time": 360, "temperature_setpoint": 21.0},
            ],
            "sunday": [
                {"start_time": 1200, "temperature_setpoint": 20.0},
            ],
        }
    )
    assert time_program.sunday[0].end_time == 1440 + 360

    time_program = RoomTimeProgram.from_api(
        **{
            "monday": [
                {"start_time": 360, "setpoint": 21.0},
            ]
        }
    )
    assert time_program.monday[0].temperature_setpoint == 21.0

    with pytest.raises(ValueError):
        RoomTimeProgram.from_api(
            **{
                "monday": [
                    {"start_time": 360, "end_time": 480, "setpoint": 21.0},
                ]
            }
        )



#!/usr/bin/env python3

import argparse
import asyncio
import copy
import hashlib
import json
import logging
import secrets
import signal
from typing import Any, TypeVar

import sys
from datetime import datetime, timedelta
from pathlib import Path
from urllib.parse import urlencode

from myPyllant.api import get_api_base, get_system_api_base
from myPyllant.const import DEFAULT_CONTROL_IDENTIFIER
from myPyllant.utils import add_default_parser_args

logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(
    description="Generates test data necessary to run integration tests."
)
add_default_parser_args(parser)
parser.add_argument(
    "--debug", help="Print debug information", action=argparse.BooleanOptionalAction
)

SALT = secrets.token_bytes(16)
DATA_DIR = Path(__file__).resolve().parent / "data"
ANONYMIZE_ATTRIBUTES = (
    "device_uuid",
    "device_serial_number",
    "deviceId",
    "homeName",
    "serialNumber",
    "systemId",
)


def user_json_dir(user: str) -> Path:
    return (
        DATA_DIR
        / hashlib.sha1(user.encode("UTF-8") + SALT, usedforsecurity=False).hexdigest()
    )


async def main(user, password, brand, country=None, write_results=True):
    """
    Generate json data for running testcases.

    Parameters:
        user (str): myVAILLANT app username
        password (str): myVAILLANT app password
        brand (str): Brand, such as vaillant or bulex
        country (str): myVAILLANT app country
        write_results (bool): Whether to write results to disk, or return them
    """
    from myPyllant.api import MyPyllantAPI
    from myPyllant.enums import DeviceDataBucketResolution
    from myPyllant.utils import datetime_format

    results: dict[str, Any] = {}
    json_dir = user_json_dir(user)
    if write_results:
        json_dir.mkdir(parents=True, exist_ok=True)

    def create_result(result, name, directory=None):
        if write_results:
            if directory:
                write_dir = json_dir / directory
            else:
                write_dir = json_dir
            with open(write_dir / f"{name}.json", "w") as fh_json:
                fh_json.write(json.dumps(result, indent=2))
        if directory:
            if directory not in results:
                results[directory] = {}
            results[directory][name] = result
        else:
            results[name] = result

    async with MyPyllantAPI(user, password, brand, country) as api:
        homes_url = f"{get_api_base()}/homes"
        async with api.aiohttp_session.get(
            homes_url, headers=api.get_authorized_headers()
        ) as homes_resp:
            homes = await homes_resp.json()
            anonymized_homes = _recursive_data_anonymize(copy.deepcopy(homes), SALT)
            for home in anonymized_homes:
                if "address" in home:
                    home.pop("address")
            create_result(
                anonymized_homes,
                name="homes",
            )

        if not homes:
            # No homes means no systems to generate test data for
            print("No homes found.")
            if write_results:
                print(f"Wrote homes.json to {json_dir}")
                exit(0)
            else:
                return results

        for home in homes:
            anonymized_home = _recursive_data_anonymize(copy.deepcopy(home), SALT)
            anonymized_system_id = anonymized_home["systemId"]
            real_system_id = home["systemId"]
            print(f"Generating test data for {anonymized_system_id}...")

            if write_results:
                (json_dir / anonymized_system_id).mkdir(parents=True, exist_ok=True)

            control_identifier_url = f"{get_api_base()}/systems/{real_system_id}/meta-info/control-identifier"
            try:
                async with api.aiohttp_session.get(
                    control_identifier_url, headers=api.get_authorized_headers()
                ) as ci_response:
                    control_identifier_json = await ci_response.json()
                    control_identifier = control_identifier_json["controlIdentifier"]
                    create_result(
                        control_identifier_json,
                        "control_identifier",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(
                    f"Error fetching {control_identifier_url}: {e}", exc_info=e
                )
                control_identifier = DEFAULT_CONTROL_IDENTIFIER

            tz_url = f"{get_api_base()}/systems/{real_system_id}/meta-info/time-zone"
            try:
                async with api.aiohttp_session.get(
                    tz_url, headers=api.get_authorized_headers()
                ) as tz_response:
                    create_result(
                        await tz_response.json(),
                        "time_zone",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {tz_url}: {e}", exc_info=e)

            dtc_url = (
                f"{get_api_base()}/systems/{real_system_id}/diagnostic-trouble-codes"
            )
            try:
                async with api.aiohttp_session.get(
                    dtc_url, headers=api.get_authorized_headers()
                ) as dtc_response:
                    dtc = await dtc_response.json()
                    anonymized_dtc = _recursive_data_anonymize(copy.deepcopy(dtc), SALT)
                    create_result(
                        anonymized_dtc,
                        "diagnostic_trouble_codes",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {dtc_url}: {e}", exc_info=e)

            try:
                connection_status_url = f"{get_api_base()}/systems/{real_system_id}/meta-info/connection-status"
                async with api.aiohttp_session.get(
                    connection_status_url, headers=api.get_authorized_headers()
                ) as status_resp:
                    create_result(
                        await status_resp.json(),
                        "connection_status",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {connection_status_url}: {e}", exc_info=e)

            system_url = get_system_api_base(real_system_id, control_identifier)
            try:
                async with api.aiohttp_session.get(
                    system_url, headers=api.get_authorized_headers()
                ) as system_resp:
                    system = await system_resp.json()
                    anonymized_homes = _recursive_data_anonymize(
                        copy.deepcopy(system), SALT
                    )
                    create_result(
                        anonymized_homes,
                        "system",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {system_url}: {e}", exc_info=e)

            current_system_url = (
                f"{get_api_base()}/emf/v2/{real_system_id}/currentSystem"
            )
            try:
                async with api.aiohttp_session.get(
                    current_system_url, headers=api.get_authorized_headers()
                ) as current_system_resp:
                    current_system = await current_system_resp.json()
                    anonymized_current_system = _recursive_data_anonymize(
                        copy.deepcopy(current_system), SALT
                    )
                    create_result(
                        anonymized_current_system,
                        "current_system",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {current_system_url}: {e}", exc_info=e)

            mpc_url = f"{get_api_base()}/hem/{real_system_id}/mpc"
            try:
                async with api.aiohttp_session.get(
                    mpc_url, headers=api.get_authorized_headers()
                ) as mpc_resp:
                    mpc = await mpc_resp.json()
                    anonymized_mpc = _recursive_data_anonymize(copy.deepcopy(mpc), SALT)
                    create_result(
                        anonymized_mpc,
                        "mpc",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {mpc_url}: {e}", exc_info=e)

            rts_url = f"{get_api_base()}/rts/{real_system_id}/devices"
            try:
                async with api.aiohttp_session.get(
                    rts_url, headers=api.get_authorized_headers()
                ) as rts_resp:
                    rts = await rts_resp.json()
                    anonymized_rts = _recursive_data_anonymize(copy.deepcopy(rts), SALT)
                    create_result(
                        anonymized_rts,
                        "rts",
                        anonymized_system_id,
                    )
            except Exception as e:
                logger.error(f"Error fetching {rts_url}: {e}", exc_info=e)

            device = current_system["primary_heat_generator"]
            start = datetime.now().replace(
                microsecond=0, second=0, minute=0, hour=0
            ) - timedelta(days=1)
            end = datetime.now().replace(microsecond=0, second=0, minute=0, hour=0)
            if device and "data" in device:
                querystring = {
                    "resolution": DeviceDataBucketResolution.HOUR,
                    "operationMode": device["data"][0]["operation_mode"],
                    "energyType": device["data"][0]["value_type"],
                    "startDate": datetime_format(start),
                    "endDate": datetime_format(end),
                }
                device_buckets_url = (
                    f"{get_api_base()}/emf/v2/{real_system_id}/"
                    f"devices/{device['device_uuid']}/buckets?{urlencode(querystring)}"
                )
                try:
                    async with api.aiohttp_session.get(
                        device_buckets_url, headers=api.get_authorized_headers()
                    ) as device_buckets_resp:
                        device_buckets = await device_buckets_resp.json()
                    create_result(
                        device_buckets,
                        "device_buckets",
                        anonymized_system_id,
                    )
                except Exception as e:
                    logger.error(
                        f"Error fetching {device_buckets_url}: {e}", exc_info=e
                    )

        rbr_url = (
            f"{get_api_base()}/api/v1/ambisense/facilities/{real_system_id}/capability"
        )
        try:
            async with api.aiohttp_session.get(
                rbr_url, headers=api.get_authorized_headers()
            ) as rbr_resp:
                rbr = await rbr_resp.json()
                create_result(
                    rbr,
                    "rbr",
                    anonymized_system_id,
                )
        except Exception as e:
            logger.error(f"Error fetching {rbr_url}: {e}", exc_info=e)

        rooms_url = (
            f"{get_api_base()}/api/v1/ambisense/facilities/{real_system_id}/rooms"
        )
        try:
            async with api.aiohttp_session.get(
                rooms_url, headers=api.get_authorized_headers()
            ) as rooms_resp:
                rooms = await rooms_resp.json()
                anonymized_rooms = _recursive_data_anonymize(copy.deepcopy(rooms), SALT)
                create_result(
                    anonymized_rooms,
                    "rooms",
                    anonymized_system_id,
                )
        except Exception as e:
            logger.error(f"Error fetching {rooms_url}: {e}", exc_info=e)

        energy_management_url = (
            f"{get_api_base()}/eebus/energy-management/{real_system_id}"
        )
        try:
            async with api.aiohttp_session.get(
                energy_management_url, headers=api.get_authorized_headers()
            ) as energy_management_resp:
                energy_management = await energy_management_resp.json()
                create_result(
                    energy_management,
                    "energy_management",
                    anonymized_system_id,
                )
        except Exception as e:
            logger.error(f"Error fetching {energy_management_url}: {e}", exc_info=e)

        ship_self_url = f"{get_api_base()}/ship/{real_system_id}/self"
        try:
            async with api.aiohttp_session.get(
                ship_self_url, headers=api.get_authorized_headers()
            ) as ship_self_resp:
                ship_self = await ship_self_resp.json()
                create_result(
                    ship_self,
                    "ship_self",
                    anonymized_system_id,
                )
        except Exception as e:
            logger.error(f"Error fetching {ship_self_url}: {e}", exc_info=e)

    if write_results:
        print(f"Wrote test data to {json_dir}")
        print()
        print(
            "Results are already anonymized, but the myVAILLANT API may change and contain new, sensitive data."
        )
        print("You should check the contents of the JSON files before sharing them.")
    else:
        return results


T = TypeVar("T")


def _recursive_data_anonymize(data: T, salt: bytes = b"") -> T:
    if isinstance(data, list):
        for elem in data:
            _recursive_data_anonymize(elem, salt)

    elif isinstance(data, dict):
        for elem in data.keys():
            if elem in ANONYMIZE_ATTRIBUTES:
                data[elem] = hashlib.sha1(
                    data[elem].encode("UTF-8") + salt, usedforsecurity=False
                ).hexdigest()
                continue
            _recursive_data_anonymize(data[elem], salt)

    return data


def signal_handler(user, sig):
    user_json_dir(user).rmdir()
    sys.exit(sig)


if __name__ == "__main__":
    args = parser.parse_args()
    if args.debug:
        logging.basicConfig(level="DEBUG")

    signal.signal(signal.SIGINT, lambda sig, _: signal_handler(args.user, sig))

    asyncio.run(main(args.user, args.password, args.brand, args.country))



import argparse
import base64
import hashlib
import random
import re
import string
from datetime import datetime, tzinfo, timedelta
from enum import Enum

from myPyllant.const import BRANDS, COUNTRIES, DEFAULT_BRAND, DEFAULT_HOLIDAY_DURATION


def dict_to_snake_case(d):
    """
    Convert {'camelCase': value} to {'camel_case': value} recursively
    """
    if d is None:
        return None

    snake_pattern = re.compile(r"(?<!^)(?=[A-Z])")

    def to_snake(s):
        return snake_pattern.sub("_", s).lower()

    if isinstance(d, list):
        return [dict_to_snake_case(i) if isinstance(i, (dict, list)) else i for i in d]
    return {
        to_snake(a): dict_to_snake_case(b) if isinstance(b, (dict, list)) else b
        for a, b in d.items()
    }


def dict_to_camel_case(d):
    """
    Convert {'camel_case': value} to {'camelCase': value} recursively
    """
    if d is None:
        return None

    def to_camel(s):
        p = s.split("_")
        return p[0] + "".join(x.capitalize() or "_" for x in p[1:])

    if isinstance(d, list):
        return [dict_to_camel_case(i) if isinstance(i, (dict, list)) else i for i in d]
    return {
        to_camel(a): dict_to_camel_case(b) if isinstance(b, (dict, list)) else b
        for a, b in d.items()
    }


def random_string(length):
    return "".join(
        random.choices(
            string.ascii_lowercase + string.ascii_uppercase + string.digits, k=length
        )
    )


def generate_code() -> tuple[str, str]:
    rand = random.SystemRandom()
    code_verifier = "".join(rand.choices(string.ascii_letters + string.digits, k=128))

    code_sha_256 = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    b64 = base64.urlsafe_b64encode(code_sha_256)
    code_challenge = b64.decode("utf-8").replace("=", "")

    return code_verifier, code_challenge


def datetime_format(date: datetime, with_microseconds=False) -> str:
    """
    Some endpoints need three digits of microseconds, some don't :shrug:
    """
    if with_microseconds:
        return date.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    else:
        return date.strftime("%Y-%m-%dT%H:%M:%S") + "Z"


def datetime_parse(date_string: str, timezone: tzinfo) -> datetime:
    if "." in date_string:
        return datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ").replace(
            tzinfo=timezone
        )
    else:
        return datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=timezone
        )


def get_realm(brand: str, country: str | None = None) -> str:
    """
    Vaillant and SDBG use `brand-country-b2c` as the realm, Bulex & Glow-worm don't use a country
    """
    match brand:
        case "bulex" | "glow-worm":
            return f"{brand}-b2c"
        case _ if not country:
            return f"{brand}-b2c"
        case _:
            return f"{brand}-{country}-b2c"


def add_default_parser_args(parser: argparse.ArgumentParser):
    parser.add_argument("user", help="Username (email address) for the myVaillant app")
    parser.add_argument("password", help="Password for the myVaillant app")
    parser.add_argument(
        "brand",
        help="Brand your account is registered in, i.e. 'vaillant'",
        default=DEFAULT_BRAND,
        choices=BRANDS.keys(),
    )
    parser.add_argument(
        "--country",
        help="Country your account is registered in, i.e. 'germany'",
        choices=COUNTRIES[DEFAULT_BRAND].keys(),
        required=False,
    )


def version_tuple(v):
    # See https://stackoverflow.com/a/11887825
    return tuple(map(int, (v.split("."))))


def prepare_field_value_for_dict(value):
    from myPyllant.models import MyPyllantDataClass

    match value:
        case tzinfo() | Enum():
            value = str(value)
        case MyPyllantDataClass():
            value = value.prepare_dict()
        case dict():
            value = {k: prepare_field_value_for_dict(v) for k, v in value.items()}
        case list():
            value = [prepare_field_value_for_dict(v) for v in value]
    return value


def get_default_holiday_dates(
    start, end, timezone, default_holiday_duration=DEFAULT_HOLIDAY_DURATION
) -> tuple[datetime, datetime]:
    """
    If no start or end date is given, use the current date and add the default holiday duration
    """
    if not start:
        start = datetime.now(timezone)
    if not end:
        end = start + timedelta(days=default_holiday_duration)
    return start, end


def recursive_compare(d1, d2, level="root"):
    """
    See https://stackoverflow.com/a/53818532
    """
    if isinstance(d1, dict) and isinstance(d2, dict):
        if d1.keys() != d2.keys():
            s1 = set(d1.keys())
            s2 = set(d2.keys())
            print("    {:<20} + {} - {}".format(level, s1 - s2, s2 - s1))
            common_keys = s1 & s2
        else:
            common_keys = set(d1.keys())

        for k in common_keys:
            recursive_compare(d1[k], d2[k], level="{}.{}".format(level, k))

    elif isinstance(d1, list) and isinstance(d2, list):
        if len(d1) != len(d2):
            print("    {:<20} len1={}; len2={}".format(level, len(d1), len(d2)))
        common_len = min(len(d1), len(d2))

        for i in range(common_len):
            recursive_compare(d1[i], d2[i], level="{}[{}]".format(level, i))

    else:
        if d1 != d2:
            print("    {:<20} {} != {}".format(level, d1, d2))



from .generate_test_data import _recursive_data_anonymize


def test_anonymize_systems_data():
    original_systems_data = [
        {
            "gateway": {
                "deviceId": "PRIVATE_deviceId",
                "serialNumber": "PRIVATE_serialNumber",
                "systemId": "PRIVATE_systemId",
                "diagnosticTroubleCodes": [],
            },
            "devices": [
                {
                    "deviceId": "PRIVATE_deviceId",
                    "serialNumber": "PRIVATE_serialNumber",
                    "articleNumber": "0020260951",
                    "name": "sensoHOME",
                    "type": "CONTROL",
                    "systemId": "PRIVATE_systemId",
                    "diagnosticTroubleCodes": [],
                },
                {
                    "deviceId": "PRIVATE_deviceId",
                    "serialNumber": "PRIVATE_serialNumber",
                    "articleNumber": "0010022040",
                    "name": "ecoTEC",
                    "type": "HEAT_GENERATOR",
                    "operationalData": {},
                    "systemId": "PRIVATE_systemId",
                    "diagnosticTroubleCodes": [],
                    "properties": [],
                },
            ],
            "status": {},
            "systemControlState": {},
            "systemConfiguration": {},
            "systemId": "PRIVATE_systemId",
            "hasOwnership": True,
        }
    ]

    anonymized_data = _recursive_data_anonymize(original_systems_data)

    assert anonymized_data == [
        {
            "gateway": {
                "deviceId": "707e295cf73b49584939e3fe6e0a96713ca22800",
                "serialNumber": "f18bbf5a183946e978558a1556ed62c4cfb9dc5e",
                "systemId": "823c180c5c91bff1ab96e14d253aaa8218b511f4",
                "diagnosticTroubleCodes": [],
            },
            "devices": [
                {
                    "deviceId": "707e295cf73b49584939e3fe6e0a96713ca22800",
                    "serialNumber": "f18bbf5a183946e978558a1556ed62c4cfb9dc5e",
                    "articleNumber": "0020260951",
                    "name": "sensoHOME",
                    "type": "CONTROL",
                    "systemId": "823c180c5c91bff1ab96e14d253aaa8218b511f4",
                    "diagnosticTroubleCodes": [],
                },
                {
                    "deviceId": "707e295cf73b49584939e3fe6e0a96713ca22800",
                    "serialNumber": "f18bbf5a183946e978558a1556ed62c4cfb9dc5e",
                    "articleNumber": "0010022040",
                    "name": "ecoTEC",
                    "type": "HEAT_GENERATOR",
                    "operationalData": {},
                    "systemId": "823c180c5c91bff1ab96e14d253aaa8218b511f4",
                    "diagnosticTroubleCodes": [],
                    "properties": [],
                },
            ],
            "status": {},
            "systemControlState": {},
            "systemConfiguration": {},
            "systemId": "823c180c5c91bff1ab96e14d253aaa8218b511f4",
            "hasOwnership": True,
        }
    ]


def test_anonymize_current_system():
    original_current_system_data = {
        "system_type": "BOILER_OR_ELECTRIC_HEATER",
        "primary_heat_generator": {
            "device_uuid": "private_device_uuid",
            "ebus_id": "BAI00",
            "spn": 376,
            "bus_coupler_address": 0,
            "article_number": "0010022040",
            "emfValid": True,
            "device_serial_number": "private_device_serial_number",
            "device_type": "BOILER",
            "first_data": "2022-11-10T13:33:34Z",
            "last_data": "2023-02-28T13:20:31Z",
            "data": [],
            "product_name": "VMW 23CS/1-5 C (N-ES) ecoTEC plus",
        },
        "secondary_heat_generators": [],
        "electric_backup_heater": None,
        "solar_station": None,
        "ventilation": None,
    }

    anonymized_data = _recursive_data_anonymize(original_current_system_data)

    assert anonymized_data == {
        "system_type": "BOILER_OR_ELECTRIC_HEATER",
        "primary_heat_generator": {
            "device_uuid": "53be2a18985b87f5613e32a9ba44cc4ccbb4f6b4",
            "ebus_id": "BAI00",
            "spn": 376,
            "bus_coupler_address": 0,
            "article_number": "0010022040",
            "emfValid": True,
            "device_serial_number": "d97b53961ee397e2e9b1bb0cabff02949c360f4f",
            "device_type": "BOILER",
            "first_data": "2022-11-10T13:33:34Z",
            "last_data": "2023-02-28T13:20:31Z",
            "data": [],
            "product_name": "VMW 23CS/1-5 C (N-ES) ecoTEC plus",
        },
        "secondary_heat_generators": [],
        "electric_backup_heater": None,
        "solar_station": None,
        "ventilation": None,
    }
